/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#ifdef __GNUG__
#pragma implementation
int dummy_GenericMapper_cc = 0;
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "GenericMapper.h"

#include <math.h>
#include <ctype.h>
#include <time.h>
#include <string>
#include <strstream.h>
#include "NetObject.h"
#include "Natto.h"
#include "GenericMapperBox.h"

#define HOURS  (60 * 60)
#define DAYS   (24 * HOURS)
#define YEARS  (365 * DAYS)


char* GenericMapper::dimension_names[n_dimensions] =
{
    "xaxis", "yaxis", "zaxis", "radius",
};

char* GenericMapper::parameter_names[n_parameters] =
{
    "profile", "degree_of_interest", "spring", "file_name",
    "content_length", "last_modified",
    "ratio_of_japanese", "n_images", "n_links",
    "file_name_135", "file_name_246",
};

char* GenericMapper::dimension_descriptions[n_dimensions] =
{
    "X-axis", "Y-axis" "Z-axis", "Size",
};

char* GenericMapper::parameter_descriptions[n_parameters] =
{
    "Profile", "Degree of Interest", "Spring", "File/Directory Name",
    "File Size (bytes)", "Last Update (days ago)",
    "% of Japanese (chars)", "Number of Images", "Number of Links",
    "File Name (1,3...)", "File Name (2,4...)",
};

GenericMapper::GenericMapper() :
    Mapper("X", "Y"),
    box_(NULL), upto_(time(NULL)), since_(upto_ - 1 * YEARS)
{
    mapping_[xaxis]  = file_name_135;
    mapping_[yaxis]  = file_name_246;
    mapping_[zaxis]  = degree_of_interest;
    mapping_[radius] = n_links;
    this->set_title();
}


void
GenericMapper::configure(
    Natto*     natto,
    GuiObject* frame)
{
    if (!box_) box_ = new GenericMapperBox("mapper", frame, this);
    box_->create(NULL);
    box_->popup(natto);
}


void
GenericMapper::set_title()
{
    this->x_title(parameter_descriptions[mapping_[xaxis]]);
    this->y_title(parameter_descriptions[mapping_[yaxis]]);
}



GenericMapper::parameter_t*
GenericMapper::mapping()
{
    return mapping_;
}


void
GenericMapper::mapping(
    parameter_t map[n_dimensions])
{
    memcpy(mapping_, map, sizeof(mapping_));
}


float
GenericMapper::map_file_name(
    Natto::node* vp)
{
    string file = vp->url().basename();
    float  value = 0;
    bool   valid = false;

    for (unsigned i = 0, j = i; i < file.length(); i++) {
	int ch = toupper(file[i]);
	if (isalpha(ch) || isdigit(ch)) {
	    if (isdigit(ch)) ch = 'Z' + 1;
	    value += (ch - 'A') * pow(26, - (double) j++);
	    valid = true;
	}
    }
    return valid ? value / 27.0: 1.0;
}


float
GenericMapper::map_file_name_135(
    Natto::node* vp)
{
    string file = vp->url().basename();
    float  value = 0;
    bool   valid = false;

    for (unsigned i = 0, j = 0; i < file.length(); i += 2) {
	int ch = toupper(file[i]);
	if (isalpha(ch) || isdigit(ch)) {
	    if (isdigit(ch)) ch = 'Z' + 1;
	    value += (ch - 'A') * pow(26, - (double) j++);
	    valid = true;
	}
    }
    return valid ? value / 27.0: 1.0;
}


float
GenericMapper::map_file_name_246(
    Natto::node* vp)
{
    string file = vp->url().basename();
    float  value = 0;
    bool   valid = false;

    for (unsigned i = 1, j = 0; i < file.length(); i += 2) {
	int ch = toupper(file[i]);
	if (isalpha(ch) || isdigit(ch)) {
	    if (isdigit(ch)) ch = 'Z' + 1;
	    value += (ch - 'A') * pow(26, - (double) j++);
	    valid = true;
	}
    }
    return valid ? value / 27.0: 1.0;
}


bool
GenericMapper::label_alphabetical(
    int     nth,
    float&  pos,
    string& label)
{
    if (nth < 0 || 13 < nth) {
	pos = 0.0;
	label = "";
	return false;
    }
    nth *= 2;
    pos = nth / 27.0;
    if (nth == 26) label = "0-9";
    else           label = 'A' + nth;
    return true;
}


float
GenericMapper::map_content_length(
    Natto::node* vp)
{
    float value = 0;
    value = vp->netobj()->content_length() / 3000.0;
    if (value > 1.0) value = 1.0;
    return value;
}


bool
GenericMapper::label_content_length(
    int     nth,
    float&  pos,
    string& label)
{
    if (nth < 0 || 15 < nth) {
	pos = 0.0;
	label = "";
	return false;
    }
    nth *= 200;
    pos = nth / 3000.0;
    strstream buf;
    buf << nth << ends;
    label = buf.str();
    buf.freeze(false);
    return true;
}


float
GenericMapper::map_last_modified(
    Natto::node* vp)
{
    float value = 0;

    if (vp->netobj()->last_modified() > 0) {
	value =
	    (float)(vp->netobj()->last_modified() - since_) / (upto_ - since_);
	if      (value < 0) value = 0.0;
	else if (value > 1) value = 1.0;
	value = pow(value, 10);
    }
    else if (vp->in_size()) {
	for (Natto::link_iter i = vp->in_begin(); i != vp->in_end(); i++)
            value += (*i)->from()->x;
	value /= vp->in_size();
    }
    else {
	value = 0.5;
    }
    return value;
}


bool
GenericMapper::label_last_modified(
    int     nth,
    float&  pos,
    string& label)
{
    time_t grid[] = {
	365 * DAYS,
	 80 * DAYS,
	 60 * DAYS,
	 50 * DAYS,
	 40 * DAYS,
	 30 * DAYS,
	 20 * DAYS,
	 15 * DAYS,
	 10 * DAYS,
	  7 * DAYS,
	  5 * DAYS,
	  2 * DAYS,
	  0 * DAYS,
    };
    if (nth < 0 || sizeof(grid)/sizeof(grid[0]) < (unsigned) nth) {
	pos = 0.0;
	label = "";
	return false;
    }
    time_t t = upto_ - grid[nth];
    pos = pow((double)(t - since_) / (upto_ - since_), 10);
    strstream buf;
    buf << (upto_ - t) / DAYS << ends;
    label = buf.str();
    buf.freeze(false);
    return true;
}


float
GenericMapper::map_ratio_of_japanese(
    Natto::node*)
{
#if 0
    NattoNode& info = vp->info();
    int total = info.n_j_letters + info.n_e_letters;
    if (total) return (float) info.n_j_letters / total;
    else       return 0.5;
#else
    return 0.5;
#endif
}


bool
GenericMapper::label_ratio_of_japanese(
    int     nth,
    float&  pos,
    string& label)
{
    if (nth < 0 || 10 < nth) {
	pos = 0.0;
	label = "";
	return false;
    }
    nth *= 10;
    pos = nth / 100.0;
    strstream buf;
    buf << nth << '%' << ends;
    label = buf.str();
    buf.freeze(false);
    return true;
}


float
GenericMapper::map_n_images(
    Natto::node*)
{
#if 0
    NattoNode& info = vp->info();
    float value = info.n_images / 10.0;
    if (value > 1.0) value = 1.0;
    return value;
#else
    return 0.5;
#endif
}


bool
GenericMapper::label_n_images(
    int     nth,
    float&  pos,
    string& label)
{
    if (nth < 0 || 10 < nth) {
	pos = 0.0;
	label = "";
	return false;
    }
    pos = nth / 10.0;
    strstream buf;
    buf << nth << ends;
    label = buf.str();
    buf.freeze(false);
    return true;
}


float
GenericMapper::map_n_links(
    Natto::node* vp)
{
    float value = (vp->in_size() + vp->out_size()) / 30.0;
    if (value > 1.0) value = 1.0;
    return value;
}


bool
GenericMapper::label_n_links(
    int     nth,
    float&  pos,
    string& label)
{
    if (nth < 0 || 15 < nth) {
	pos = 0.0;
	label = "";
	return false;
    }
    nth *= 2;
    pos = nth / 30.0;
    strstream buf;
    buf << nth << ends;
    label = buf.str();
    buf.freeze(false);
    return true;
}


float
GenericMapper::map_by_parameter(
    Natto::node* vp,
    parameter_t para)
{
    switch (para) {
     case profile:           return 0;
     case file_name:         return map_file_name(vp);
     case file_name_135:     return map_file_name_135(vp);
     case file_name_246:     return map_file_name_246(vp);
     case content_length:    return map_content_length(vp);
     case last_modified:     return map_last_modified(vp);
     case ratio_of_japanese: return map_ratio_of_japanese(vp);
     case n_images:          return map_n_images(vp);
     case n_links:           return map_n_links(vp);
     default:                return 0;
    }
    return 0;
}


bool
GenericMapper::label_by_parameter(
    int     nth,
    float&  pos,
    string& label,
    parameter_t para)
{
    switch (para) {
     case profile:           return 0;
     case file_name: case file_name_135: case file_name_246:
			     return label_alphabetical(nth, pos, label);
     case content_length:    return label_content_length(nth, pos, label);
     case last_modified:     return label_last_modified(nth, pos, label);
     case ratio_of_japanese: return label_ratio_of_japanese(nth, pos, label);
     case n_images:          return label_n_images(nth, pos, label);
     case n_links:           return label_n_links(nth, pos, label);
     default:                return 0;
    }
    return 0;
}


void
GenericMapper::layout(
    Natto*, Natto::node* vp)
{
    if (!vp) return;
    
    vp->x  = map_by_parameter(vp, mapping_[xaxis]);
    vp->y  = map_by_parameter(vp, mapping_[yaxis]);
    vp->z  = z_step() * vp->doi() + log(vp->out_size() + 1) * 0.02;
    float r = def_radius() * log(vp->in_size() + vp->out_size() + 2);
    if (vp->r != r) {
	vp->r = r;
	vp->set_flags(Natto::shape_changed_bit);
    }
}


bool
GenericMapper::x_label(
    Natto*,
    int     nth,
    float&  pos,
    string& label)
{
    return label_by_parameter(nth, pos, label, mapping_[xaxis]);
}


bool
GenericMapper::y_label(
    Natto*,
    int     nth,
    float&  pos,
    string& label)
{
    return label_by_parameter(nth, pos, label, mapping_[yaxis]);
}
