/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#define DBG_MODULE  Natto

#ifdef __GNUG__
#pragma implementation
int dummy_Natto_cc = 0;
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "Natto.h"

#include <algorithm>
#include "util.h"
#include "dfs.h"
#include "NetObject.h"
#include "GenericMapper.h"
#include "NattoView.h"


NattoNode::NattoNode():
    natto_(NULL), netobj_(NULL), shape_(NULL), doi_(0),
    x(0), y(0), z(0), r(0), dx(0), dy(0), dz(0),
    flags_(0), visited_(0), written_(0)
{}


NattoNode::NattoNode(
    NetObject* netobj,
    Natto*     natto):
    natto_(natto), netobj_(netobj), shape_(NULL), doi_(0),
    x(0), y(0), z(0), r(0), dx(0), dy(0), dz(0),
    flags_(0), visited_(0), written_(0)
{
    // DBG("node", url() << " creating");
    if (netobj_) netobj_->attach(this);
}


NattoNode::NattoNode(
    const NattoNode& a):
    natto_(a.natto_), netobj_(a.netobj_), shape_(a.shape_), doi_(a.doi_),
    x(a.x), y(a.y), z(a.z), r(a.r), dx(a.dx), dy(a.dy), dz(a.dz),
    flags_(a.flags_), visited_(a.visited_), written_(a.written_)
{
    if (netobj_) netobj_->attach(this);
}


NattoNode::~NattoNode()
{
    // DBG("node", url() << " destroying");
    if (netobj_) netobj_->detach(this);
    delete shape_;
}


void
NattoNode::natto(
    Natto* n)
{
    natto_ = n;
}


void
NattoNode::shape(
    NodeShape* s)
{
    shape_ = s;
}


string
NattoNode::label(
    int len)
{
    string s =
        netobj_->title().size() ? netobj_->title(): netobj_->url().basename();
    return len ? s.substr(0, len): s;
}


URL
NattoNode::url() const
{
    return netobj_->url();
}


void
NattoNode::doi(
    float x)
{
    doi_ = x;
}


void
NattoNode::set_flags(
    int bits)
{
    flags_ |= bits;
}


void
NattoNode::unset_flags(
    int bits)
{
    flags_ &= ~bits;
}


void
NattoNode::set_visited(
    int bits)
{
    visited_ |= bits;
}


void
NattoNode::unset_visited(
    int bits)
{
    visited_ &= ~bits;
}


void
NattoNode::update(observable*, int, void*)
{
    natto_->node_scanned(DCAST(Natto::node*, this));
}




NattoLink::NattoLink():
    natto_(NULL), netanchor_(NULL), shape_(NULL), visited_(0)
{}


NattoLink::NattoLink(
    NetAnchor* netanchor,
    Natto*     natto):
    natto_(natto), netanchor_(netanchor), shape_(NULL),
    flags_(0), visited_(0), distance_(0)
{
    // DBG("link", href() << " creating");
}


NattoLink::NattoLink(
    const NattoLink& r):
    natto_(r.natto_), netanchor_(r.netanchor_), shape_(r.shape_),
    flags_(r.flags_), visited_(r.visited_), distance_(r.distance_)
{}


NattoLink::~NattoLink()
{
    // DBG("link", href() << " destroying");
    delete shape_;
}


void
NattoLink::natto(
    Natto* n)
{
    natto_ = n;
}


void
NattoLink::shape(
    LinkShape* s)
{
    shape_ = s;
}


string
NattoLink::href() const
{
    return netanchor_->href();
}


string
NattoLink::label(
    int len)
{
    string s = netanchor_->anchor().size() ? netanchor_->anchor(): href();
    return len ? s.substr(0, len): s;
}


void
NattoLink::distance(
    float d)
{
    distance_ = d;
}


void
NattoLink::set_flags(
    int bits)
{
    flags_ |= bits;
}


void
NattoLink::unset_flags(
    int bits)
{
    flags_ &= ~bits;
}


void
NattoLink::set_visited(
    int bits)
{
    visited_ |= bits;
}


void
NattoLink::unset_visited(
    int bits)
{
    visited_ &= ~bits;
}



Natto::Natto():
    graph_(new graph_type),
    active_node_(NULL), active_link_(NULL),
    mapper_(new GenericMapper)
{}


Natto::~Natto()
{
    delete graph_;
    delete mapper_;
}


Natto::node*
Natto::find_node(
    URL  url,
    bool insert_mode)
{
    DBG("node", "finding (and inserting) " << url);
    url.frag("");
    for (node_iter vi  = this->nodes_begin(); vi != this->nodes_end(); vi++)
        if ((*vi)->url() == url) return *vi;
    return insert_mode ? this->insert_node(url): (node*) 0;
}


Natto::node*
Natto::insert_node(
    URL url)
{
    NetObject* netobj = NetObject::find_by_url(url);
    if (!netobj) netobj = new NetObject(url);
    return this->insert_node(NattoNode(netobj, this));
}


Natto::node*
Natto::insert_node(
    const NattoNode& info)
{
    node* vp = graph_->insert(info);
    vp->natto(this);
    this->changed(true);
    return vp;
}


Natto::link*
Natto::insert_link(
    const NattoLink& info,
    node*            from,
    node*            to)
{
    link* ep = graph_->insert(info, from, to);
    ep->natto(this);
    this->changed(true);
    return ep;
}


void
Natto::erase_node(
    node* vp)
{
    graph_->erase(vp);
    this->changed(true);
}


void
Natto::erase_link(
    link* ep)
{
    graph_->erase(ep);
    this->changed(true);
}


void
Natto::activate_node(
    node* vp)
{
    if (active_node_ == vp) return;
    DBG("act", "node=" << (vp ? vp->url().abs(): string("(none)")));
    this->activate_link(NULL);

    if (active_node_) {
        active_node_->unset_flags(active_bit);
        active_node_->set_flags(shape_changed_bit);
        link_iter ei  = active_node_->out_begin();
        link_iter end = active_node_->out_end();
        for (; ei != end; ei++) {
            (*ei)->unset_flags(active_bit);
            (*ei)->set_flags(shape_changed_bit);
        }
    }

    this->changed(true);
    active_node_ = vp;

    if (active_node_) {
        active_node_->set_flags(shape_changed_bit | active_bit);
        for (link_iter ei  = active_node_->out_begin();
             ei != active_node_->out_end(); ei++) {
            (*ei)->set_flags(shape_changed_bit | active_bit);
            vp = (*ei)->to();
            if (vp->doi() < active_node_->doi() - 1)
                vp->doi(active_node_->doi() - 1);
            this->mapper()->layout(this, vp);
        }
        this->mapper()->layout(this, active_node_);
        this->scan_node(active_node_, false);
    }
}


void
Natto::activate_link(
    link* ep)
{
    if (active_link_ == ep) return;
    DBG("act", "link=" << (ep ? ep->href(): string("(none)")));

    if (active_link_) {
        active_link_->unset_flags(active_child_bit);
        active_link_->set_flags(shape_changed_bit);
        node* vp = active_link_->to();
        vp->unset_flags(active_child_bit);
        vp->set_flags(shape_changed_bit);
        for (link_iter ei = vp->out_begin(); ei != vp->out_end(); ei++) {
            (*ei)->unset_flags(active_link_bit);
            (*ei)->set_flags(shape_changed_bit);
        }
    }

    active_link_ = ep;
    this->changed(true);
    
    if (active_link_) {
        active_link_->set_flags(shape_changed_bit | active_child_bit);
        node* vp = active_link_->to();
        vp->set_flags(shape_changed_bit | active_child_bit);
        for (link_iter ei  = vp->out_begin(); ei != vp->out_end(); ei++)
            (*ei)->set_flags(shape_changed_bit | active_link_bit);
        this->scan_node(vp, false);
    }
}


void
Natto::scan_node(
    node* vp,
    bool  rescan)
{
    if (rescan || !vp->test_flags(Natto::scanned_bit))
        vp->netobj()->download(true);
}


void
Natto::node_scanned(
    node* vp)
{
    NetObject* netobj = vp->netobj();
    if (netobj->is_redirect()) {
        netobj->detach(vp);
        netobj = vp->netobj()->location();
        netobj->attach(vp);
        netobj->download(true);
        return;
    }

    vp->set_flags(Natto::scanned_bit);
    for (Natto::link_riter ei = vp->out_rbegin(); ei != vp->out_rend(); ei++)
        this->erase_link(*ei);

    if (netobj->is_error()) {
        vp->set_flags(Natto::disabled_bit);
        for (Natto::link_riter ei = vp->in_rbegin();
             ei != vp->in_rend(); ei++) {
            // this->erase_link(*ei);
            (*ei)->set_flags(Natto::disabled_bit);
        }
        this->sync();
        return;
    }
    vp->unset_flags(Natto::disabled_bit);

    NetObject::anchor_iter i   = vp->netobj()->anchors_begin();
    NetObject::anchor_iter end = vp->netobj()->anchors_end();
    for (; i != end; i++) {
        Natto::node* up = this->find_node(URL((*i).href(), vp->url()), true);
        Natto::link* ep = this->insert_link(NattoLink(&*i, this), vp, up);

        ep->unset_flags(Natto::disabled_bit | Natto::masked_bit);
        int f = 0;
        if ((f = vp->test_flags(Natto::disabled_bit | Natto::masked_bit) |
                 up->test_flags(Natto::disabled_bit | Natto::masked_bit))) 
            ep->set_flags(f);

        if (vp == this->active_node())
            ep->set_flags(Natto::active_bit | Natto::shape_changed_bit);

        if (vp->url().host() == up->url().host()) {
            string relative = ep->href();
            unsigned distance = 1;
            for (string::size_type j = 0; j < relative.length(); j++)
                if (relative[j] == '/') distance++;
            ep->distance(distance);
        }
        else {
            ep->distance(10);
        }
        if (up->doi() < vp->doi() - 1) up->doi(vp->doi() - 1);
        this->mapper()->layout(this, vp);
        this->mapper()->layout(this, up);
    }
    
    this->sync();
    return;
}


void
Natto::clear_visited()
{
    for (node_iter vi  = this->nodes_begin(); vi != this->nodes_end(); vi++)
        (*vi)->unset_visited(1);
}


void
Natto::lift_up(
    node* vp)
{
    if (!vp) vp = active_node_;
    if (!vp) return;
    ArrangeNodes ar_up(this, ArrangeNodes::up, vp->doi());
    this->clear_visited();
    (void) dfs_forward(vp, ar_up, (int) vp->doi() + 1);
    this->changed(true);
}


void
Natto::lift_down(
    node* vp)
{
    if (!vp) vp = active_node_;
    if (!vp) return;
    ArrangeNodes ar_down(this, ArrangeNodes::down, vp->doi());
    this->clear_visited();
    (void) dfs_forward(vp, ar_down);
    this->changed(true);
}


void
Natto::lift_down_all()
{
    for (node_iter vi = this->nodes_begin();  vi != this->nodes_end(); vi++) {
	(*vi)->doi(0);
	mapper()->layout(this, *vi);
    }
    this->changed(true);
}


void
Natto::set_mark(
    string mark,
    node*  vp)
{
    if (!vp) return;
    mark_map::iterator i = marks_.find(mark);
    if (i != marks_.end()) {
        node* up = (*i).second;
	vector<string>::iterator j = 
	    find(up->mark_keys.begin(), up->mark_keys.end(), mark);
	if (j != up->mark_keys.end()) up->mark_keys.erase(j);
    }
    marks_.insert(mark_map::value_type(mark, vp));
    vp->mark_keys.push_back(mark);
    vp->set_flags(shape_changed_bit);
    this->changed(true);
}


Natto::node*
Natto::get_mark(
    string mark)
{
    mark_map::iterator i = marks().find(mark);
    if (i == marks_.end()) return NULL;
    return (*i).second;
}


int
ArrangeNodes::operator()(
    Natto::node* vp,
    Natto::link*,
    int depth)
{
    float level = vp->doi();
    int result = (level < 1) ? dfs_visited: dfs_next;

    if (vp->is_visited(1) && dir_ == down) {
	return dfs_visited;
    }
    if (dir_ == up   && level > first_level_ - depth || 
	dir_ == down && level < first_level_ - depth) {
	vp->set_visited(1);
	return dfs_next;
    }
    vp->set_visited(1);
    if (dir_ == up)     vp->doi(++level);
    else if (level > 0) vp->doi(--level);
    
    natto_->mapper()->layout(natto_, vp);
    DBG("lift", "level=" << level << ' ' << vp->url());
    return result;
}
