/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#ifndef NATTO_H
#define NATTO_H
#ifdef __GNUG__
#pragma interface
#endif

#include <string>
#include <map>
#include <list>
#include <istream.h>
#include <ostream.h>
#include "utilstl.h"
#include "graph.h"
//#include "dfs.h"
#include "observer.h"
#include "URL.h"

class NetObject;
class NetAnchor;
class NodeShape;
class LinkShape;
class Natto;
class Mapper;


class NattoNode: public observer
{
 public:
    NattoNode();
    NattoNode(NetObject*, Natto*);
    NattoNode(URL, Natto*);
    NattoNode(const NattoNode&);
    virtual ~NattoNode();

    Natto*     natto()  { return natto_; }
    NetObject* netobj() { return netobj_; }
    NodeShape* shape()  { return shape_; }
    
    string label(int = 0);
    URL    url() const;
    float  doi() const { return doi_; }
    void   doi(float);
    
    int  test_flags(int bits) const { return flags_ & bits; }
    void set_flags(int bits);
    void unset_flags(int bits);

    int  is_visited(int bits) { return visited_ & bits; }
    void set_visited(int bits);
    void unset_visited(int bits);

    void update(observable*, int, void*);

    vector<string> mark_keys;

    void natto(Natto*);
    void shape(NodeShape*);

 private:
    Natto*     natto_;
    NetObject* netobj_;
    NodeShape* shape_;

    float  doi_;

 public:
    float x, y, z, r;
    float dx, dy, dz; // spring mapper

 private:
    int flags_;
    int visited_;
    int written_;  // serial number in writing

    NattoNode& operator=(const NattoNode&);
    DBG_NEW
};



class NattoLink
{
 public:
    NattoLink();
    NattoLink(NetAnchor*, Natto*);
    NattoLink(const NattoLink&);
    virtual ~NattoLink();

    Natto*     natto()     { return natto_; }
    NetAnchor* netanchor() { return netanchor_; }
    LinkShape* shape()     { return shape_; }

    string label(int = 0);
    string href() const;
    float  distance() const { return distance_; }
    
    int  test_flags(int bits) const { return flags_ & bits; }
    void set_flags(int bits);
    void unset_flags(int bits);

    int  is_visited(int bits) { return visited_ & bits; }
    void set_visited(int bits);
    void unset_visited(int bits);

    void natto(Natto*);
    void shape(LinkShape*);
    void distance(float);

 private:
    Natto*     natto_;
    NetAnchor* netanchor_;
    LinkShape* shape_;

    int    flags_;
    int    visited_;
    float  distance_;

    NattoLink& operator=(const NattoLink&);
    DBG_NEW
};



class Natto: public observable
{
 public:
    enum priority_t { pri_ignorable, pri_sync };
    enum flag_bits_t {
	scanned_bit  = 0x001, active_bit = 0x002, marked_bit = 0x004,
	disabled_bit = 0x008, masked_bit = 0x010, drawn_bit  = 0x020,
        active_child_bit  = 0x040, active_link_bit = 0x080,
        shape_changed_bit = 0x100
    };

    typedef graph<NattoNode, NattoLink> graph_type;
    
    // types
    typedef graph_type::vertex node;
    typedef graph_type::edge   link;
    typedef graph_type::vertex_container node_container;
    typedef graph_type::edge_container   link_container;
    typedef graph_type::vertex_iter  node_iter;
    typedef graph_type::edge_iter    link_iter;
    typedef graph_type::vertex_riter node_riter;
    typedef graph_type::edge_riter   link_riter;
    typedef smap<string, Natto::node*> mark_map;

    Natto();
    virtual ~Natto();
    
    // graph management
    node* insert_node(URL);
    node* insert_node(const NattoNode& info);
    link* insert_link(const NattoLink& info, node* from, node* to);
    void  erase_node(node*);
    void  erase_link(link*);
    void  activate_node(node*);
    void  activate_link(link*);
    void  scan_node(node* vp, bool rescan);
    node* find_node(URL url, bool insertion = true);
    void  node_scanned(node*);

    node* active_node() { return active_node_; }
    link* active_link() { return active_link_; }
    
    // lift-up and -down operations
    void lift_up(node* = NULL);
    void lift_down(node* = NULL);
    void lift_down_all();

    // mark management
    mark_map& marks() { return marks_; }  // #####
    void      set_mark(string, node*);
    void      set_mark(string mark) { this->set_mark(mark, active_node_); }
    node*     get_mark(string);

    void sync() { this->notify(pri_sync, 0); }

    node_iter nodes_begin() { return graph_->verteces_begin(); }
    node_iter nodes_end()   { return graph_->verteces_end(); }
    link_iter links_begin() { return graph_->edges_begin(); }
    link_iter links_end()   { return graph_->edges_end(); }
    node_container::size_type nodes_size() { return graph_->verteces_size(); }
    link_container::size_type links_size() { return graph_->edges_size(); }
    
    void clear_visited();

    // member access
    Mapper* mapper() const    { return mapper_; }
    void    mapper(Mapper* m) { mapper_ = m; }

 private:
    graph_type* graph_;        // representaion
    node*       active_node_;  // active node
    link*       active_link_;  // active link
    mark_map    marks_;        // mark-onto-node map
    Mapper*     mapper_;

    Natto(const Natto&);
    Natto& operator=(const Natto&);
    DBG_NEW
};



class ArrangeNodes
{
    Natto*  natto_;
    int     dir_;
    float   first_level_;
 public:
    enum direction_t { up, down };

    ArrangeNodes(Natto* natto, direction_t dir, float level):
    natto_(natto), dir_(dir), first_level_(level) {}
    int operator()(Natto::node* vp, Natto::link* ep, int depth);

 private:
    ArrangeNodes(const ArrangeNodes&);
    ArrangeNodes& operator=(const ArrangeNodes&);
    DBG_NEW
};

#endif /* NATTO_H */
