/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#define DBG_MODULE  View

#ifdef __GNUG__
#pragma implementation
int dummy_NattoView_cc = 0;
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "NattoView.h"

#include <math.h>
#include <queue>
#include <stdlib.h>
#include <string.h>
#include "utilgl.h"
#include "Mapper.h"
#include "TourBox.h"
#include "rgb.h"


#define INC_ROT      10.0
#define INC_TRANS     0.1


NattoView::NattoView(
    const char* name,
    GuiObject*  parent,
    bool        install_cmap,
    bool        double_buffer,
    MainFrame*  frame):

    GuiObject(name, parent),
    install_cmap_(install_cmap), double_buffer_(double_buffer),
    frame_(frame), rootview_(NULL),
    x_(0), y_(0), width_(1), height_(1),
    render_mode_(draw_label_bit),
    z_board_(- 2.0 * natto()->mapper()->def_radius()),
    inc_trans_(INC_TRANS), inc_rot_(INC_ROT),
    level_(0), tourbox_(NULL)
{
    this->init_colors();
    rot_    = coord(-80.0,  0.0,  0.0);
    trans_  = coord(  0.0, -0.4, -2.1);
    origin_ = coord(  0.5,  0.5,  0.0);
}


NattoView::NattoView(
    const char* name,
    GuiObject*  parent,
    NattoView*  view,
    int         x,
    int         y,
    int         width,
    int         height):

    GuiObject(name, parent),
    install_cmap_(view->install_cmap()), double_buffer_(view->double_buffer()),
    frame_(view->frame()), rootview_(frame_->rootview()),
    x_(x), y_(y), width_(width), height_(height),
    render_mode_(view->render_mode_),
    z_board_(view->z_board_),
    inc_trans_(view->inc_trans_), inc_rot_(view->inc_rot_),
    level_(view->level_ + 1)
{
    this->init_colors();
}


NattoView::~NattoView()
{
    delete tourbox_;
    this->close_all_subviews();
}


void
NattoView::init_colors()
{
#define COLOR3(c, r, g, b)     c[0] = r, c[1] = g, c[2] = b
#define COLOR4(c, r, g, b, a)  c[0] = r, c[1] = g, c[2] = b, c[3] = a
#define COLOR3X(c, macro)      COLOR3(c, macro)
#define COLOR4X(c, macro, a)   COLOR4(c, macro, a)

    COLOR4(light_ambient,  0.2, 0.2,  0.2, 1.0);
    COLOR4(light_diffuse,  0.8, 0.8,  0.8, 1.0);
    COLOR4(light_specular, 0.1, 0.1,  0.1, 1.0);
    COLOR4(light_position, 0.0, 0.0, 10.0, 0.0);
    light_shininess  = 10.0;

    //COLOR4(material_ambient,  0.4, 0.4, 0.4, 1.0);
    COLOR4(material_diffuse,  0.8, 0.8, 0.8, 1.0);
    COLOR4(material_specular, 0.8, 0.8, 0.8, 1.0);

    COLOR4(background0_color, 0.2, 0.2, 0.5, 0.0);
    COLOR4(background1_color, 0.0, 0.0, 0.0, 0.0);

    COLOR3(cursor_color, 0.0, 0.7, 1.0);
    COLOR3(grid_color,   0.8, 0.8, 0.8);

    COLOR3(board1_color, 0.2, 0.2, 0.5);
    COLOR3(board2_color, 0.6, 0.2, 0.2);
    COLOR3(board3_color, 0.7, 0.5, 0.1);
    COLOR3(board4_color, 0.5, 0.7, 0.2);

    COLOR3(normal_node_color,  1.0, 0.8, 0.0);
    COLOR3(scanned_node_color, 0.8, 0.6, 0.0);
    COLOR3(active_node_color,  1.0, 0.4, 0.0);
    COLOR3(marked_node_color,  0.4, 0.7, 0.0);

    COLOR4(normal_link1_color, 0.8, 1.0, 1.0, 0.5);
    COLOR4(normal_link2_color, 1.0, 0.5, 0.5, 0.3);
    COLOR4(active_link1_color, 0.4, 1.0, 0.0, 0.6);
    COLOR4(active_link2_color, 0.4, 1.0, 0.0, 0.6);
    //COLOR4(path_link_color,    0.8, 0.8, 0.8, 0.5);

    COLOR3(normal_text_color,  0.8, 0.8, 0.8);
    COLOR3(strong_text_color,  0.8, 1.0, 1.0);
    COLOR3(popup_text_color,   1.0, 1.0, 1.0);
    COLOR3(popup_plate_color,  0.3, 0.3, 0.6);
    COLOR3(active_plate_color, 0.5, 0.3, 0.0);
    COLOR3(marked_plate_color, 0.1, 0.4, 0.5);
    //COLOR3(status_text_color,  1.0, 1.0, 1.0);

    normal_link_width = 1.0;
#ifdef MESA
    active_link_width = 1.5;
    normal_text_width = 1.5;
    cursor_width      = 1.5;
#else
    active_link_width = 1.0;
    normal_text_width = 1.0;
    cursor_width      = 1.0; 
#endif

#undef COLOR3
#undef COLOR4
#undef COLOR3X
#undef COLOR4X
}


NattoView_if*
NattoView::interface()
{
    return DCAST(NattoView_if*, imp()->interface());
}


void
NattoView::create(
    GuiConfig* config)
{
    GuiObject::create(&PortKit::create_NattoView, config);
}


void
NattoView::update()
{
    DBG("draw", name() << " updating");

    subviews_type::iterator i;

    this->update_prefix();
    for (i = subviews_.begin(); i != subviews_.end(); i++)
	(*i)->update_prefix();

    this->draw();
    for (i = subviews_.begin(); i != subviews_.end(); i++)
	(*i)->draw();

    interface()->make_current();
    interface()->swap_buffers();
    for (i = subviews_.begin(); i != subviews_.end(); i++) {
	(*i)->interface()->make_current();
	(*i)->interface()->swap_buffers();
    }

    interface()->make_current();
    interface()->adjust_sliders();
    for (i = subviews_.begin(); i != subviews_.end(); i++) {
	(*i)->interface()->make_current();
	(*i)->interface()->adjust_sliders();
    }
}


void
NattoView::render_wireframe()
{
    this->render_mode(wireframe_bit, 0);
}


void
NattoView::render_surface()
{
    this->render_mode(draw_label_bit, wireframe_bit | draw_icon_bit);
}


void
NattoView::render_draw_icon()
{
    this->render_mode(draw_icon_bit, wireframe_bit);
}


void
NattoView::render_no_label()
{
    this->render_mode(draw_label_bit, 0);
}


void
NattoView::draw()
{
    this->draw_prefix();
    this->draw_board();
    this->draw_nodes();
    this->draw_cursor();
    // draw_links() must be done at last! because of blending.
    this->draw_links();
    this->draw_suffix();
}


void
NattoView::draw_nodes()
{
    this->draw_nodes_prefix();
    Natto::node_iter begin = natto()->nodes_begin();
    Natto::node_iter end   = natto()->nodes_end();
    for (node_drawing_iter vi(begin, end); vi != end; vi++) {
	Natto::node* vp = *vi;
	vp->set_flags(Natto::drawn_bit);
	this->draw_a_node(vp);
	vp->unset_flags(Natto::shape_changed_bit);
    }
    this->draw_nodes_suffix();
}


void
NattoView::draw_links()
{
    this->draw_links_prefix();
    Natto::link_iter begin = natto()->links_begin();
    Natto::link_iter end   = natto()->links_end();
    for (link_drawing_iter ei(begin, end); ei != end; ei++) {
	Natto::link* ep = *ei;
        ep->set_flags(Natto::drawn_bit);
	this->draw_a_link(ep);
        ep->unset_flags(Natto::shape_changed_bit);
    }
    this->draw_links_suffix();
}


void
NattoView::render_mode(
    int on_bits,
    int off_bits)
{
    render_mode_ &= ~off_bits;
    render_mode_ |= on_bits;
    Natto::node_iter vi  = natto()->nodes_begin();
    Natto::node_iter end = natto()->nodes_begin();
    for (; vi != end; vi++) (*vi)->set_flags(Natto::shape_changed_bit);
}


void
NattoView::translate(
    GLfloat& axis,
    GLfloat  inc)
{
    axis += inc;
    this->close_all_subviews();
}


void 
NattoView::rotate(
    GLfloat& axis,
    GLfloat  inc)
{
    axis += inc;
    if      (axis < -180) axis += 360;
    else if (axis >  180) axis -= 360;
    this->close_all_subviews();
}


void 
NattoView::trans_left_right_abs(
    GLfloat dist)
{
    trans_.x = dist;
    this->close_all_subviews();
}


void 
NattoView::trans_up_down_abs(
    GLfloat dist)
{
    trans_.y = dist;
    this->close_all_subviews();
}


void 
NattoView::rot_left_right_abs(
    GLfloat angle)
{
    rot_.z = angle;
    this->close_all_subviews();
}


void 
NattoView::rot_up_down_abs(
    GLfloat angle)
{
    rot_.x = angle;
    this->close_all_subviews();
}


void
NattoView::home_view()
{
    rot_    = coord(-80.0,  0.0,  0.0);
    trans_  = coord(  0.0, -0.4, -2.1);
    origin_ = coord(  0.5,  0.5,  0.0);
    this->close_all_subviews();
    this->render_mode(0, ortho_proj_bit);
}


void
NattoView::top_view()
{
    rot_    = coord(0.0, 0.0,  0.0);
    trans_  = coord(0.0, 0.0, -2.0);
    origin_ = coord(0.5 ,0.5,  0.0);
    this->close_all_subviews();
    this->render_mode(0, ortho_proj_bit);
}


void
NattoView::top_view_2d()
{
    rot_    = coord(0.0, 0.0,  0.0);
    trans_  = coord(0.0, 0.0, -2.0);
    origin_ = coord(0.5 ,0.5,  0.0);
    this->close_all_subviews();
    this->render_mode(ortho_proj_bit, 0);
}


NattoView*
NattoView::open_subview(
    const char* name,
    int         x,
    int         y,
    int         width,
    int         height)
{
    NattoView* view = interface()->open_subview(name, x, y, width, height);
    subviews_.push_back(view);
    view->set_origin();
    return view;
}


void
NattoView::set_origin()
{
    vector<Natto::node*> nodes;
    this->viewable_nodes(nodes);
#if 1
    if (nodes.size()) {
	float x = 0.5;
	float y = 0.5;
	float z = FLT_MIN;
	Natto::node* active_node = natto()->active_node();
	for (vector<Natto::node*>::iterator i = nodes.begin();
	     i != nodes.end(); i++) {
	    if (*i == active_node || (*i)->z >= z) {
		x = (*i)->x;
		y = (*i)->y;
		z = (*i)->z;
		if (*i == active_node) break;
	    }
	}
	origin_.x = x - (rootview() ? rootview()->origin().x: 0);
	origin_.y = y - (rootview() ? rootview()->origin().y: 0);
	origin_.z = z - (rootview() ? rootview()->origin().z: 0);
    }
#else
    if (nodes.size()) {
	float xmax = FLT_MIN, xmin = FLT_MAX;
	float ymax = FLT_MIN, ymin = FLT_MAX;
	float zmax = FLT_MIN, zmin = FLT_MAX;
	for (vector<Natto::node*>::iterator i = nodes.begin();
	     i != nodes.end(); i++) {
	    xmax = max(xmax, vp->x);  xmin = min(xmin, vp->x);
	    ymax = max(ymax, vp->y);  ymin = min(ymin, vp->y);
	    zmax = max(zmax, vp->z);  zmin = min(zmin, vp->z);
	}
	origin_.x = (xmax + xmin)/2 - (rootview() ? rootview()->origin().x: 0);
	origin_.y = (ymax + ymin)/2 - (rootview() ? rootview()->origin().y: 0);
	origin_.z = (zmax + zmin)/2 - (rootview() ? rootview()->origin().z: 0);
    }
#endif
}


void
NattoView::close_subview(
    NattoView* view)
{
    subviews_type::iterator i = find(subviews_.begin(), subviews_.end(), view);
    if (i != subviews_.end()) subviews_.erase(i);
    delete view;
}


void
NattoView::close_all_subviews()
{
    while (subviews_.size()) {
	delete subviews_.back();
	subviews_.pop_back();
    }
}


void
NattoView::close()
{

    if (rootview()) rootview()->close_subview(this);
    else            ExitNattoView();
}


void
NattoView::popup_tourbox()
{
    tourbox_->popup();
}


NattoView_if::~NattoView_if()
{}
