/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#ifndef NATTOVIEW_H
#define NATTOVIEW_H
#ifdef __GNUG__
#pragma interface
#endif

#include "Natto.h"
#include "MainFrame.h"

class  TourBox;
struct NodeShape { virtual ~NodeShape() {} };
struct LinkShape { virtual ~LinkShape() {} };

//////////////////////////////////////////////////////////////////////////////
//  NattoView --
//
class NattoView_if: public GuiObject_if
{
 public:
    virtual ~NattoView_if();
    
    virtual void  make_current() = 0;
    virtual void  swap_buffers() = 0;
    virtual void* string_bitmap(string str, int& w, int& h, int& base) = 0;
    virtual void  adjust_sliders() = 0;

    virtual NattoView* open_subview(const char* name,
                                    int x, int y, int w, int h) = 0;
    DBG_NEW
};


class NattoView: public GuiObject
{
 public:
    struct coord {
	float x, y, z;
	coord(float a = 0, float b = 0, float c = 0): x(a), y(b), z(c) {}
    };
    
    NattoView(const char* name, GuiObject* parent,
	      bool install_cmap, bool double_buffer, MainFrame* frame);
    NattoView(const char* name, GuiObject* parent,
	      NattoView* view, int x, int y, int width, int height);
    virtual ~NattoView();

    // as GuiObject
    virtual NattoView_if* interface();
    virtual void          create(GuiConfig*);

    // operations and manipulations to the focused node
    void focus(URL url)  { frame()->focus_node(url); }
    void lift_up()       { frame()->lift_up(); }
    void lift_down()     { frame()->lift_down(); }
    void lift_down_all() { frame()->lift_down_all(); }
    void browse()        { frame()->browse(); }
    void reload()        { frame()->scan(); }

#if 0
    // mark operations
    void set_mark(string mark, URL url) {
	natto()->set_mark(mark, natto()->find(url));
    }
    void set_mark(string mark) {
	natto()->set_mark(mark, natto()->active_node());
    }
    void set_mark() { set_mark("system"); }
#endif

    // translation operations
    void trans_right(float d) { translate(trans_.x,  d); }
    void trans_right()        { translate(trans_.x,  inc_trans_); }
    void trans_left (float d) { translate(trans_.x, -d); }
    void trans_left ()        { translate(trans_.x, -inc_trans_); }
    void trans_up   (float d) { translate(trans_.y,  d); }
    void trans_up   ()        { translate(trans_.y,  inc_trans_); }
    void trans_down (float d) { translate(trans_.y, -d); }
    void trans_down ()        { translate(trans_.y, -inc_trans_); }

    // rotation operations
    void rot_down (float a) { rotate(rot_.x,  a); }
    void rot_down ()        { rotate(rot_.x,  inc_rot_); }
    void rot_up   (float a) { rotate(rot_.x, -a); }
    void rot_up   ()        { rotate(rot_.x, -inc_rot_); }
    void rot_right(float a) { rotate(rot_.z,  a); }
    void rot_right()        { rotate(rot_.z,  inc_rot_); }
    void rot_left (float a) { rotate(rot_.z, -a); }
    void rot_left ()        { rotate(rot_.z, -inc_rot_); }

    // zoom operatoins
    void zoom_in (float d) { translate(trans_.z,  d); }
    void zoom_in ()        { translate(trans_.z,  inc_trans_); }
    void zoom_out(float d) { translate(trans_.z, -d); }
    void zoom_out()        { translate(trans_.z, -inc_trans_); }

    void trans_left_right_abs(float);
    void trans_up_down_abs(float);
    void rot_left_right_abs(float);
    void rot_up_down_abs(float);
    
    // rendering mode
    enum render_t {
	wireframe_bit  = 1, draw_icon_bit  = 2,
	draw_label_bit = 4, ortho_proj_bit = 8,
    };
    int  render_mode() const { return render_mode_; }
    void render_mode(int on_bits, int off_bits);
    void render_wireframe();
    void render_surface();
    void render_draw_icon();
    void render_no_label();

    // view control
    void home_view();
    void top_view();
    void top_view_2d();
    void focus_view() {}

    // picking objects
    virtual void pick(int x, int y, vector<int>& sysobjs,
		      vector<Natto::node*>& nodes,
                      vector<Natto::link*>& links) =0;
    virtual void viewable_nodes(vector<Natto::node*>&) = 0;

    // update 
    void update();

    // functions for debug and evaluation
    virtual void get_raster_pos(Natto::node* v, float& x, float& y) = 0;

    NattoView* open_subview(const char*, int x, int y, int width, int height);
    void       close_subview(NattoView* view);
    void       close_all_subviews();
    void       close();

    // member access
    MainFrame* frame()    { return frame_; }
    NattoView* rootview() { return rootview_; }
    Natto*     natto()    { return frame()->natto(); }

    int   xpos()      const { return x_; }
    int   ypos()      const { return y_; }
    int   width()     const { return width_; }
    int   height()    const { return height_; }
    float z_board()   const { return z_board_; }
    coord trans()     const { return trans_; }
    coord rot()       const { return rot_; }
    coord origin()    const { return origin_; }
    float inc_trans() const { return inc_trans_; }
    float inc_rot()   const { return inc_rot_; }
    int   level()     const { return level_; }
    bool  double_buffer() { return double_buffer_; }
    void  double_buffer(bool f) { double_buffer_ = f; }
    bool  install_cmap() { return install_cmap_; }
    void  install_cmap(bool f) { install_cmap_ = f; }

    void popup_tourbox();
    virtual void init_graphics() = 0;

// protected:
    void init_colors();

    // subroutines of update()
    virtual void update_prefix() {}
    void         draw();

    // subroutines of draw()
    virtual void draw_prefix() {}
    virtual void draw_suffix() {}
    virtual void draw_board() {}
    virtual void draw_cursor() {}

    void         draw_nodes();
    virtual void draw_nodes_prefix() {}
    virtual void draw_nodes_suffix() {}
    virtual void draw_a_node(Natto::node*) = 0;

    void         draw_links();
    virtual void draw_links_prefix() {}
    virtual void draw_links_suffix() {}
    virtual void draw_a_link(Natto::link*) = 0;

    // subroutines of trans_*() and rot_*()
    void translate(float& axis, float inc);
    void rotate(float& axis, float inc);

    void    set_origin();

    virtual void resize(int x, int y, int w, int h) {
	x_ = x;
	y_ = y;
	width_ = w;
	height_ = h;
    }

    TourBox* tourbox()            { return tourbox_; }
    void     tourbox(TourBox* tb) { tourbox_ = tb; }

 private:
    bool       install_cmap_;
    bool       double_buffer_;
    MainFrame* frame_;
    NattoView* rootview_;
    int        x_;
    int        y_;
    int        width_;
    int        height_;
    int        render_mode_;
    float      z_board_;
    coord      trans_;
    coord      rot_;
    coord      origin_;
    float      inc_trans_;
    float      inc_rot_;
    int        level_;

    typedef vector<NattoView*> subviews_type; 
    subviews_type subviews_;
    TourBox*  tourbox_;

 protected:
    float light_ambient[4];
    float light_diffuse[4];
    float light_specular[4];
    float light_position[4];
    float light_shininess;

    float material_diffuse[4];
    float material_specular[4];
    
    float background0_color[4];
    float background1_color[4];

    float cursor_color[3];
    float grid_color[3];

    float board1_color[3];
    float board2_color[3];
    float board3_color[3];
    float board4_color[3];
    
    float normal_node_color[3];
    float scanned_node_color[3];
    float active_node_color[3];
    float marked_node_color[3];
    
    float normal_link2_color[4];
    float normal_link1_color[4];
    float active_link1_color[4];
    float active_link2_color[4];
    
    float normal_text_color[3];
    float strong_text_color[3];
    float popup_text_color[3];
    float popup_plate_color[3];
    float active_plate_color[3];
    float marked_plate_color[3];
    
    float normal_link_width;
    float active_link_width;
    float normal_text_width;
    float cursor_width;

 private:
    NattoView(const NattoView&);
    NattoView& operator=(const NattoView&);
    DBG_NEW
};


class node_drawing_iter
{
 public:
    node_drawing_iter(Natto::node_iter iter, Natto::node_iter end):
        iter_(iter), end_(end) { adjust(); }
    
    Natto::node* operator*() {
        return *iter_;
    }
    
    node_drawing_iter operator++(int) {
        node_drawing_iter prev = *this;
        iter_++;
        adjust();
        return prev;
    }

    node_drawing_iter operator++() {
        iter_++;
        adjust();
        return *this;
    }

    bool operator!=(Natto::node_iter i) {
        return iter_ != i;
    }

 protected:
    void adjust() {
        while (iter_ != end_ &&
               (*iter_)->test_flags(Natto::disabled_bit | Natto::masked_bit))
            (*iter_++)->unset_flags(Natto::drawn_bit);
    }
    
 private:
    Natto::node_iter iter_;
    Natto::node_iter end_;
    DBG_NEW
};


class link_drawing_iter
{
 public:
    link_drawing_iter(Natto::link_iter iter, Natto::link_iter end):
        iter_(iter), end_(end) { adjust(); }
    
    Natto::link* operator*() {
        return *iter_;
    }
    
    link_drawing_iter operator++(int) {
        link_drawing_iter prev = *this;
        iter_++;
        adjust();
        return prev;
    }

    link_drawing_iter operator++() {
        iter_++;
        adjust();
        return *this;
    }

    bool operator!=(Natto::link_iter i) {
        return iter_ != i;
    }

 protected:
    void adjust() {
        while (iter_ != end_ &&
               (*iter_)->test_flags(Natto::disabled_bit | Natto::masked_bit))
            (*iter_++)->unset_flags(Natto::drawn_bit);
    }
    
 private:
    Natto::link_iter iter_;
    Natto::link_iter end_;
    DBG_NEW
};

#endif /* NATTOVIEW_H */
