/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#ifndef NETOBJECT_H
#define NETOBJECT_H
#ifdef __GNUG__
#pragma interface
#endif

#include <vector>
#include <string>
#include "utilstl.h"
#include "observer.h"
//#include "www.h"
#include "URL.h"

class NetHeaders;


class NetAnchor
{
    typedef NetAnchor self_type;
    
 public:
    enum link_type_t {
	unknown_link, normal_link,
	form_get, form_post, map_area, frame_src, img_src, img_dynsrc,
	body_background, bgsound, input_src, script_src,
        rel_next, rel_prev, rel_made, rel_stylesheet, rel_script, rel_print
    };

    NetAnchor();
    NetAnchor(string href, link_type_t ltype = normal_link,
              string anchor = "", string title = "", string mtype = "",
              link_type_t rev = unknown_link);
    virtual ~NetAnchor();

    string href()    const { return href_; }
    string anchor()  const { return anchor_; }

    bool operator==(const NetAnchor& r) const { return href_ == r.href(); }
    bool operator< (const NetAnchor& r) const { return href_ <  r.href(); }
    
 private:
    string      href_;
    link_type_t link_type_;
    string      anchor_;
    string      title_;
    string      media_type_;
    link_type_t rev_;
    DBG_NEW2
};


class NetObject: public observable
{
    friend class LinkExtractor;
    
 public:
    typedef vector<NetAnchor> anchors_type;
    typedef anchors_type::iterator anchor_iter;
    typedef anchors_type::const_iterator anchor_citer;
    typedef anchors_type::reverse_iterator anchor_riter;
    typedef anchors_type::const_reverse_iterator anchor_criter;
    
    explicit NetObject(URL);
    virtual ~NetObject();

    URL    url()   const { return url_; }
    URL    base()  const { return base_ ? base_: url_; }
    string title() const { return title_; }

    int  status()      const { return status_; }
    bool is_success()  const { return status_ == 200; }
    bool is_accessed() const { return status_ != 0; }
    bool is_redirect() const { return status_ / 100 == 3; }
    bool is_error()    const { return status_ < 0 || 400 <= status_; }

    anchor_iter anchors_begin() { return anchors_.begin(); }
    anchor_iter anchors_end()   { return anchors_.end(); }

    void download(bool notifying);
    void notify() { this->observable::notify(0, NULL); }
    
    string     content_type()   const { return content_type_; }
    size_t     content_length() const { return content_length_; }
    time_t     date()           const { return date_; }
    time_t     expires()        const { return expires_; }
    time_t     last_modified()  const { return last_modified_; }
    NetObject* location()       const { return location_; }

    static NetObject* find_by_url(URL);
    static void       clear_database();

 protected:
    void base(URL);
    void title(string);
    void hyperlink(const NetAnchor&);
    
 private:
    URL         url_;
    int         status_;
    NetObject*  location_;
    URL         base_;
    string      title_;
    NetHeaders* headers_;
    anchors_type anchors_;

    void   cache_headers();
    string content_type_;
    size_t content_length_;
    time_t date_;
    time_t expires_;
    time_t last_modified_;
 
 private:
    typedef hmap<string, NetObject*> netobjs_type;
    static netobjs_type netobjs_;

    NetObject(const NetObject&);
    NetObject& operator=(NetObject&);
    DBG_NEW
};

#endif /* NETOBJECT_H */
