/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#ifdef __GNUG__
#pragma implementation
int dummy_SpringMapper_cc = 0;
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "SpringMapper.h"

#include <iomanip.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include "util.h"
#include "Natto.h"


SpringMapper::SpringMapper(
    int dimensions):
    Mapper("", ""),
    dim_(dimensions), epsilon(0.001), L(0.01), K(1), kqq(0.000001),
    nodes_size_(0)
{}


inline double
SpringMapper::alpha(
    int)
{
    return 0.1 * L;
}


inline void
SpringMapper::distance(
    const Natto::node* vp,
    const Natto::node* up,
    double& r,
    double& rx,
    double& ry,
    double& rz)
{
    double dx = dim_ & dim_x ? up->x - vp->x: 0;
    double dy = dim_ & dim_y ? up->y - vp->y: 0;
    double dz = dim_ & dim_z ? up->z - vp->z: 0;
    if (dx == 0 && dy == 0 && dz == 0) {
	r  = 0.001 * L * rand() / RAND_MAX;
	if (dim_ & dim_x) rx = 0.1 * rand() / RAND_MAX;
	if (dim_ & dim_y) ry = 0.1 * rand() / RAND_MAX;
	if (dim_ & dim_z) rz = 0.1 * rand() / RAND_MAX;
    }
    else {
	r = sqrt(dx*dx + dy*dy + dz*dz);
	if (r < 0.001 * L) r = 0.001 * L;
	if (dim_ & dim_x) rx = dx / r;
	if (dim_ & dim_y) ry = dy / r;
	if (dim_ & dim_z) rz = dz / r;
    }
}


void
SpringMapper::layout(
    Natto*       natto,
    Natto::node* vp)
{
    if (!vp) return;

    if (!(dim_ & dim_z)) 
	vp->z  = z_step() * vp->doi() + log(vp->out_size() + 1) * 0.02;

    if (!vp->test_flags(Natto::drawn_bit)) {
	vp->x = vp->y = vp->z = 0;
	if (vp->in_size()) {
	    for (Natto::link_iter ei = vp->in_begin();
                 ei != vp->in_end(); ei++) {
		Natto::node* vp_from = (*ei)->from();
		if (vp_from->test_flags(Natto::disabled_bit)) continue;
		if (dim_ & dim_x) vp->x += vp_from->x;
		if (dim_ & dim_y) vp->y += vp_from->y;
		if (dim_ & dim_z) vp->z += vp_from->z;
	    }
	    if (dim_ & dim_x)
		vp->x /= vp->in_size() + 0.2 * rand() / RAND_MAX - 0.1;
	    if (dim_ & dim_y)
		vp->y /= vp->in_size() + 0.2 * rand() / RAND_MAX - 0.1;
	    if (dim_ & dim_z)
		vp->z /= vp->in_size() + 0.2 * rand() / RAND_MAX - 0.1;
	}
	else {
	    if (dim_ & dim_x) vp->x = 0.5 * rand() / RAND_MAX + 0.25;
	    if (dim_ & dim_y) vp->y = 0.5 * rand() / RAND_MAX + 0.25;
	    if (dim_ & dim_z) vp->z = 0.5 * rand() / RAND_MAX + 0.25;
	}
    }

    float r = def_radius() * log(vp->in_size() + vp->out_size() + 2);
    if (vp->r != r) {
	vp->r = r;
	vp->set_flags(Natto::shape_changed_bit);
    }

    if (!vp->test_flags(Natto::drawn_bit)) {
	nodes_size_ = natto->nodes_size();
	this->simulate(100, natto, vp);
    }
    natto->notify(Natto::pri_ignorable, 0);
}


void
SpringMapper::layout_all(
    Natto* natto)
{
    nodes_size_ = natto->nodes_size();
    this->simulate(1000, natto);
    natto->sync();
}


bool
SpringMapper::x_label(
    Natto*, int,
    float&  pos,
    string& label)
{
    pos = 0.0;
    label = "";
    return false;
}


bool
SpringMapper::y_label(
    Natto*, int,
    float&  pos,
    string& label)
{
    pos = 0.0;
    label = "";
    return false;
}


void
SpringMapper::simulate(
    int          max_iter,
    Natto*       natto,
    Natto::node* vp)
{
    // L = 0.1 * pow(natto->nodes().size(), -0.1);

    Natto::node_iter nodes_begin = natto->nodes_begin();
    Natto::node_iter nodes_end   = natto->nodes_end();
    Natto::link_iter links_begin = natto->links_begin();
    Natto::link_iter links_end   = natto->links_end();
#if 0
    for (Natto::node_iter vi = nodes_begin; vi != nodes_end; vi++)
	(*vi)->info().dx = (*vi)->info().dy = 0;
#endif
    double delta = DBL_MAX;
    double prev_delta = DBL_MAX;
    for (int iter = 1;
	 delta > epsilon && prev_delta > epsilon && iter < max_iter; iter++) {
	prev_delta = delta;
	delta = 0.0;

	msgout << "Layout";
	if (vp) msgout << ' ' << vp->url();
	msgout << ": " << setw(3) << iter << "..." << endl;

	for (Natto::link_iter ei = links_begin; ei != links_end; ei++) {
	    float d = (*ei)->distance();
	    Natto::node* vp_from = (*ei)->from();
	    Natto::node* vp_to   = (*ei)->to();
	    if (vp_from->test_flags(Natto::disabled_bit) ||
                vp_to  ->test_flags(Natto::disabled_bit)) continue;

	    double r, rx, ry, rz;
	    distance(vp_from, vp_to, r, rx, ry, rz);
	    double dx = alpha(iter) * K * (r - L*d) * rx;
	    double dy = alpha(iter) * K * (r - L*d) * ry;
	    double dz = alpha(iter) * K * (r - L*d) * rz;

	    if (vp_to->is_visited(0)) {
		if (dim_ & dim_x) vp_from->dx += dx;
		if (dim_ & dim_y) vp_from->dy += dy;
		if (dim_ & dim_z) vp_from->dz += dz;
	    }
	    if (vp_from->is_visited(0)) {
		if (dim_ & dim_x) vp_to->dx -= dx;
		if (dim_ & dim_y) vp_to->dy -= dy;
		if (dim_ & dim_z) vp_to->dz -= dz;
	    }
	}
#if 1
	for (Natto::node_iter v1 = nodes_begin; v1 != nodes_end; v1++) {
	    Natto::node* vp1 = *v1;
	    if (vp1->test_flags(Natto::disabled_bit)) continue;

	    Natto::node_iter v2;
	    for (v2 = v1, v2++; v2 != nodes_end; v2++) {
		Natto::node* vp2 = *v2;
		if (vp2->test_flags(Natto::disabled_bit)) continue;
		
		double r, rx, ry, rz;
		distance(vp1, vp2, r, rx, ry, rz);
		double dx = - kqq / (r*r) * rx;
		double dy = - kqq / (r*r) * ry;
		double dz = - kqq / (r*r) * rz;

		if (dim_ & dim_x) vp1->dx += dx;
		if (dim_ & dim_y) vp1->dy += dy;
		if (dim_ & dim_z) vp1->dz += dz;
		if (dim_ & dim_x) vp2->dx -= dx;
		if (dim_ & dim_y) vp2->dy -= dy;
		if (dim_ & dim_z) vp2->dz -= dz;
	    }

	    double dx = vp1->dx;
	    double dy = vp1->dy;
	    double dz = vp1->dz;

	    delta += fabs(dx) + fabs(dy) + fabs(dz);
	    if (fabs(dx) > L) dx = dx > 0 ? L: -L;
	    if (fabs(dy) > L) dy = dy > 0 ? L: -L;
	    if (fabs(dz) > L) dz = dz > 0 ? L: -L;

	    if (dim_ & dim_x) vp1->x += dx;
	    if (dim_ & dim_y) vp1->y += dy;
	    if (dim_ & dim_z) vp1->z += dz;
	    if (dim_ & dim_x) vp1->dx = 0;
	    if (dim_ & dim_y) vp1->dy = 0;
	    if (dim_ & dim_z) vp1->dz = 0;
	}
#endif
	delta /= nodes_size_;
	// natto->update(NattoNotify::ignorable);
    }

    float xmax = FLT_MIN;
    float xmin = FLT_MAX;
    float ymax = FLT_MIN;
    float ymin = FLT_MAX;
    float zmax = FLT_MIN;
    float zmin = FLT_MAX;

    for (Natto::node_iter vi = nodes_begin; vi != nodes_end; vi++) {
	Natto::node* vp1 = *vi;
	if (vp1->test_flags(Natto::disabled_bit)) continue;

	if (vp1->x > xmax) xmax = vp1->x;
	if (vp1->x < xmin) xmin = vp1->x;
	if (vp1->y > ymax) ymax = vp1->y;
	if (vp1->y < ymin) ymin = vp1->y;
	if (vp1->z > zmax) zmax = vp1->z;
	if (vp1->z < zmin) zmin = vp1->z;
    }

    for (Natto::node_iter vi = nodes_begin; vi != nodes_end; vi++) {
	Natto::node* vp1 = *vi;
	if (vp1->test_flags(Natto::disabled_bit)) continue;
	if (dim_ & dim_x) vp1->x = (vp1->x - xmin) / (xmax - xmin);
	if (dim_ & dim_y) vp1->y = (vp1->y - ymin) / (ymax - ymin);
	if (dim_ & dim_z) vp1->z = (vp1->z - zmin) / (zmax - zmin);
    }
    
    msgout << "Layout";
    if (vp) msgout << ' ' << vp->url();
    msgout << ": done." << endl;
}
