/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#ifndef GRAPH_H
#define GRAPH_H
#ifdef __GNUG__
//#pragma interface
#endif

//#include <list>
#include <vector>
#include <algorithm>

//////////////////////////////////////////////////////////////////////////////
//  graph -- template class to represent graph structures
//
template <class Vertex, class Edge>
class graph
{
    typedef graph<Vertex, Edge> self_type;
 public:
    // types
    class vertex;
    class edge;
    typedef Vertex vertex_base;
    typedef Edge   edge_base;
#if 0
    typedef list<vertex*> vertex_container;
    typedef list<edge*>   edge_container;
#else
    typedef vector<vertex*> vertex_container;
    typedef vector<edge*>   edge_container;
#endif
    typedef vertex_container::iterator vertex_iter;
    typedef edge_container::iterator   edge_iter;
    typedef vertex_container::reverse_iterator vertex_riter;
    typedef edge_container::reverse_iterator   edge_riter;
    
    //////////////////////////////////////////////////////////////////////////
    //  vertex
    //
    class vertex: public vertex_base
    {
	friend class graph;
     public:
	// premitive functions
	vertex() {}
	vertex(const vertex_base& info): vertex_base(info) {}

	// member access
        edge_iter in_begin()  { return in_.begin(); }
        edge_iter in_end()    { return in_.end(); }
        edge_iter out_end()   { return out_.end(); }
        edge_iter out_begin() { return out_.begin(); }

        edge_riter in_rbegin()  { return in_.rbegin(); }
        edge_riter in_rend()    { return in_.rend(); }
        edge_riter out_rbegin() { return out_.rbegin(); }
        edge_riter out_rend()   { return out_.rend(); }

        size_t in_size()  { return in_.size(); }
        size_t out_size() { return out_.size(); }

     private:
	edge_container in_;   // incomming edges
	edge_container out_;  // outgoing edges
        DBG_NEW
    };

    //////////////////////////////////////////////////////////////////////////
    //  edge
    //
    class edge: public edge_base
    {
	friend class graph;
     public:
	// premitive functions
	edge() {}
	edge(const edge_base& info, vertex* from, vertex* to):
	    edge_base(info), from_(from), to_(to) {}
        
	// member access
	vertex* from() { return from_; }
	vertex* to()   { return to_; }

     private:
	vertex* from_;  // source vertex
	vertex* to_;    // destination vertex
        DBG_NEW
    };
    
 public:
    vertex_iter verteces_begin() { return verteces_.begin(); }
    vertex_iter verteces_end()   { return verteces_.end(); }
    edge_iter   edges_begin()    { return edges_.begin(); }
    edge_iter   edges_end()      { return edges_.end(); }
    vertex_container::size_type verteces_size() { return verteces_.size(); }
    edge_container::size_type   edges_size()    { return edges_.size(); }

    ~graph() {
        for (vertex_riter i = verteces_.rbegin(); i != verteces_.rend(); i++)
            this->erase(*i);
        for (edge_riter j = edges_.rbegin(); j != edges_.rend(); j++)
            this->erase(*j);
    }
    
    vertex* insert(const vertex_base& v) {
	vertex* vp = new vertex(v);
	verteces_.push_back(vp);
	return vp;
    }

    edge* insert(const edge_base& e, vertex* from, vertex* to) {
	edge* ep = new edge(e, from, to);
	edges_.push_back(ep);
        ep->from_ = from;
        ep->to_ = to;
	ep->from_->out_.push_back(ep);
	ep->to_->in_.push_back(ep);
	return ep;
    }

    void erase(vertex* vp) {
	vertex_iter i = find(verteces_.begin(), verteces_.end(), vp);
	if (i != verteces_.end()) verteces_.erase(i);
	while (!vp->in_.empty())  this->erase(vp->in_.back());
	while (!vp->out_.empty()) this->erase(vp->out_.back());
	delete vp;
    }

    void erase(edge* ep) {
	edge_iter i;
        i = find(edges_.begin(), edges_.end(), ep);
	if (i != edges_.end()) edges_.erase(i);
	i = find(ep->from_->out_.begin(), ep->from_->out_.end(), ep);
	if (i != ep->from_->out_.end()) ep->from_->out_.erase(i);
	i = find(ep->to_->in_.begin(), ep->to_->in_.end(), ep);
	if (i != ep->to_->in_.end()) ep->to_->in_.erase(i);
	delete ep;
    }

 private:
    vertex_container verteces_;
    edge_container   edges_;
    DBG_NEW
};

#endif /* GRAPH_H */
