/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#define DBG_MODULE  Kanji

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "kanji.h"

#include <ctype.h>
#include <string.h>
#include <strstream.h>


inline bool
is_jis(unsigned char ch) 
{
    return 0x21 <= ch && ch <= 0x7e;
}


inline bool
is_euc(unsigned char ch) 
{
    return 0xa1 <= ch && ch <= 0xfe;
}


inline bool
is_sjis1(unsigned char ch)
{
    return 0x81 <= ch && (ch <= 0x9f || 0xe0 <= ch && ch <= 0xef);
}


inline bool
is_sjis2(unsigned char ch)
{
    return 0x40 <= ch && ch <= 0xfc && ch != 0x7f;
}


int
guess_kanji_code(
    char*  s,
    size_t len,
    int    mask = KANJI_ALL)
{
    int code = KANJI_ASCII;
    unsigned char ch = '\0';
    
    if (!len) len = strlen(s);
    for (size_t i = 0; i < len && (ch = (unsigned char) *s); i++, s++) {
	if (ch == '\033') {
	    if (s[1] == '$' && strchr("@B",  s[2])) {
		code |= KANJI_JIS;
		continue;
	    }
	    if (s[1] == '(' && strchr("HBJ", s[2])) {
		code |= KANJI_JIS;
		continue;
	    }
	    continue;
	}
	if (!(ch & 0x80)) continue;
	if (is_euc(ch) && is_euc(s[1])) {
	    code |= KANJI_EUC;
	    continue;
	}
	if (is_sjis1(ch) && is_sjis2(s[1])) {
	    code |= KANJI_SJIS;
	    continue;
	}
    }
    return code & mask;
}


int
converterbuf::sync()
{
    streamsize n = pptr () - pbase ();
    for (; n > 0; n--) {
	if (putch(*pbase()) == EOF) return EOF;
    }
    return 0;
}
     

int
converterbuf::overflow(int ch)
{
    streamsize n = pptr () - pbase ();
    if (n && sync()) return EOF;
    if (ch != EOF && putch(ch) == EOF) return EOF;
    pbump(-n);
    return 0;
}


inline void
s2ebuf::convert(
    unsigned char& c1,
    unsigned char& c2)
{
    c1 <<= 1;
    if (c1 > 0x9f * 2)  { c1 += 0x80 - 0xb1 * 2 + 1; }
    else                { c1 += 0x80 - 0x71 * 2 + 1; }
    if      (c2 > 0x9e) { c2 += 0x80 - 0x7e; c1++; }
    else if (c2 > 0x7e) { c2 += 0x80 - 0x20; }
    else                { c2 += 0x80 - 0x1f; }
}


int
s2ebuf::putch(
    unsigned char ch)
{
    switch (state_) {
    case first:
	if (is_sjis1(ch)) {
	    prev_ = ch;
	    state_ = second;
	}
	else {
	    *out_ << ch;
	    state_ = first;
	}
	break;
	
    case second:
	if (is_sjis2(ch)) {
	    convert(prev_, ch);
	    *out_ << prev_ << ch;
	    state_ = first;
	}
	else {
	    *out_ << prev_ << ch;
	    state_ = first;
	}
	break;
    }
    return ch;
}


int
j2ebuf::putch(
    unsigned char ch)
{
    switch (state_) {
    case ascii:
	if (ch == '\033') {
	    state_ = in_esc;
	}
	else {
	    *out_ << ch;
	    state_ = ascii;
	}
	break;

    case in_esc:
	if (ch == '$') {
	    state_ = in_esc_dollar;
	}
	else {
	    *out_ << '\033' << ch;
	    state_ = ascii;
	}
	break;
	
    case in_esc_dollar:
	if (ch == '@' || ch == 'B') {
	    state_ = first;
	}
	else {
	    *out_ << "\033$" << ch;
	    state_ = ascii;
	}
	break;

    case first:
	if (ch == '\033') {
	    state_ = out_esc;
	}
	else if (is_jis(ch)) {
	    prev_ = ch;
	    state_ = second;
	}
	else {
	    *out_ << ch;
	    state_ = first;
	}
	break;

    case second:
	if (is_jis(ch)) {
	    *out_ << (unsigned char) (prev_ | '\200')
                  << (unsigned char) (ch    | '\200');
	    state_ = first;
	}
	else {
	    *out_ << prev_ << ch;
	    state_ = first;
	}
	break;
	
    case out_esc:
	if (ch == '(') {
	    state_ = out_esc_paren;
	}
	else {
	    *out_ << '\033';
	    state_ = first;
	    putch(ch);
	}
	break;

    case out_esc_paren:
	if (ch == 'H' || ch == 'B' || ch == 'J') {
	    state_ = ascii;
	}
	else {
	    *out_ << '\033';
	    state_ = first;
	    putch('(');
	    putch(ch);
	}
	break;
    }
    return ch;
}
    

istream&
anytoeuc(
    istream& in,
    ostream& out)
{
    char buf[1024] = "";
    size_t len = 0;

    do {
	in.get(buf, sizeof(buf), '\n');
	size_t len = in.gcount();
	int code = guess_kanji_code(buf, len, KANJI_ALL);
	ostream* str = &out;
	j2estream j2e(str); 
	if (code & KANJI_JIS) str = &j2e;
	s2estream s2e(str); 
	if (code & KANJI_SJIS) str = &s2e;
	str->write(buf, len);
    } while (sizeof(buf) == len);
    return in;
}


string
anytoeuc(
    string str)
{
    char* outbuf = new char[str.length() + 1];
    strstream in;
    in << str << ends;
    strstream out(outbuf, str.length() + 1);
    anytoeuc(in, out);
    out << ends;
    str = out.str();
    out.freeze(false);
    delete[] outbuf;
    DBG("euc", "EUC: " << str);
    return str;
}
