/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1997-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#ifdef __GNUG__
#pragma implementation
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "NetHTTP.h"

#include <strstream.h>
#include "util.h"
#include "netstream.h"

#define MODE  0666


NetHTTP::NetHTTP(
    NetManager* manager):
    NetProtocol(manager)
{}


void
NetHTTP::get(
    NetRequest&  req,
    NetResponse& res)
{
    this->do_request(req, res, "GET");
}


void
NetHTTP::head(
    NetRequest&  req,
    NetResponse& res)
{
    URL url = req.url();
    netstream* www = this->get_stream(url);
    if (!www || !www->is_open()) {
        res.status(-1);
        return;
    }
    
    RequestLine reqln("HEAD", url.reqpath(), "HTTP/1.1");
    req.headers()->set_Host(url.host());
    *www << reqln << *req.message();

    ResponseLine resln;
    NetHeaders* headers = new NetHeaders;
    *www >> resln >> *headers;
    res.message(new NetMessage(headers));

    this->release_stream(www);
}


void
NetHTTP::post(
    NetRequest&  req,
    NetResponse& res)
{
    this->do_request(req, res, "POST");
}


void
NetHTTP::do_request(
    NetRequest&  req,
    NetResponse& res,
    const char*  cmd)
{
    URL url = req.url();
    netstream* www = this->get_stream(url);
    if (!www || !www->is_open()) {
        res.status(-1);
        return;
    }
    
    RequestLine reqln(cmd, url.reqpath(), "HTTP/1.1");
    req.headers()->set_Host(url.host());
    *www << reqln << *req.message();

    ResponseLine resln;
    NetHeaders* headers = new NetHeaders;
    *www >> resln >> *headers;

    bool cache_on = !req.is_nocache();
    memobject* content = NULL;
    size_t len = 0;
    if (headers->get_Content_Length(len)) {
        if (cache_on) {
            int mode = mapped_file::in | mapped_file::out | mapped_file::trunc;
            content = new mapped_file(req.cache_file(), len, mode, MODE);
        }
        else {
            content = new heap_memory(len);
        }
        char* area = (char*) content->area();
        size_t n = 0, nread = 0;
        do {
            www->read(area + nread, len - nread);
            nread += (n = www->gcount());
        } while (*www && n && nread < len);
        len = nread;
    }
    else if (cache_on) {
        string fname = req.cache_file();
        ofstream fout(fname.c_str(), ios::out | ios::trunc, MODE);
        char buf[4096];
        size_t n = 0, nread = 0;
        do {
            www->read(buf, 4096);
            nread += (n = www->gcount());
            fout.write(buf, n);
        } while (*www && n);
        len = nread;
        fout.close();
        content = new mapped_file(fname, len, mapped_file::in, MODE);
    }
    else {
        strstream buf;
        size_t nread;
        for (nread = 0;; nread++) {
            unsigned char ch;
            www->get(ch);
            if (!*www) break;
            buf.put(ch);
        };
        len = nread;
        content = new heap_memory(buf.str(), buf.tellp());
    }

    // cache headers
    if (cache_on) {
        string meta = req.cache_file() + ".meta";
        ofstream hout(meta.c_str(), ios::out | ios::trunc);
        hout << *headers;
        hout.close();
        res.set_flags(NetResponse::cached_bit);
    }

    res.message(new NetMessage(headers, content));
    res.status(resln.status);

    this->release_stream(www);
}


ostream&
operator<<(
    ostream& out,
    const NetHTTP::RequestLine& req)
{
    out << req.method << ' ' << req.path << ' ' << req.version << "\r\n";
    return out;
}


istream&
operator>>(
    istream& in,
    NetHTTP::ResponseLine& res)
{
    strstream buf;
    if (copy_line(in, buf)) {
        buf << ends;
        buf >> res.version >> ws >> res.status >> ws;
        streampos pos = buf.tellg();
        res.description = string(buf.str(), pos);
        buf.freeze(false);
    }
    else {
        res.version = "HTTP/unknown";
        res.status = -1;
        res.description = "Response line not present.";
    }
    return in;
}


ostream&
operator<<(
    ostream& out,
    const NetHTTP::ResponseLine& res)
{
    out << res.version << ' ' << res.status << ' ' << res.description <<"\r\n";
    return out;
}
