/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1997-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#ifdef __GNUG__
#pragma implementation
int dummy_NetManager_cc = 0;
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "NetManager.h"

#include <fcntl.h>
#include <sys/types.h>
#include <limits.h>
#include <strstream.h>
#include "netstream.h"
#include "NetHTTP.h"
#include "NetLocal.h"


NetRequest::NetRequest(
    URL         url,
    method_t    method,
    int         flags,
    NetMessage* message):
    url_(url), method_(method), flags_(flags)
{
    if (message) {
        message_ = message;
        this->set_flags(freeze_bit);
    }
    else {
        message_ = new NetMessage(new NetHeaders);
    }
}


NetRequest::~NetRequest()
{
    if (!this->is_frozen()) delete message_;
}


NetResponse::NetResponse():
    status_(-1), message_(NULL), flags_(0)
{}


NetResponse::~NetResponse()
{
    if (!this->is_frozen()) delete message_;
}





NetManager::NetManager(
    string cdir)
{
    http_  = new NetHTTP(this);
    file_  = new NetLocal(this);
    if (cdir.length()   > 0) this->cache_dir(cdir);
}


NetManager::~NetManager()
{
    delete http_;
    delete file_;
}


void
NetManager::request(
    NetRequest&  req,
    NetResponse& res)
{
    time_t now = time(NULL);
    cache_data cache;
    bool use_cache = false;
    if (cache_db_) {
        if (!req.is_reloading() &&
            this->retrieve_cache(req.url(), cache) &&
            !(cache.expires != (time_t) -1 && cache.expires < now ||
             cache.access + 10 * 60 < now)) {
            req.url(string("file:") + cache_dir_ + '/' + cache.code);
            use_cache = true;
        }
        else {
            cache.code = this->cache_code(req.url());
            string path = cache_dir_ + '/' + cache.code;
            req.cache_file(path);
        }
    }
    
    string scheme = req.url().scheme();
    NetProtocol* proto = NULL;
    if      (scheme == "http")  proto = http_;
    else if (scheme == "file")  proto = file_;
    else {
        res.status(-1);
        return;
    }
    switch (req.method()) {
     case NetRequest::GET:   proto->get (req, res);  break;
     case NetRequest::POST:  proto->post(req, res);  break;
     case NetRequest::HEAD:  proto->head(req, res);  break;
    }
    if (use_cache) res.status(cache.status);
    
    if (res.is_cached()) {
        cache.status = res.status();
        NetHeaders* headers = res.headers();
        if (!headers->get_Date(cache.access)) cache.access  = now;
        if (!headers->get_Expires(cache.expires)) {
            // cgi-bin has no Last-Modified.
            time_t tmp;
            if (!headers->get_Last_Modified(tmp)) cache.expires = now;
            else  cache.expires = (time_t) -1;
        }
        this->register_cache(req.url(), cache);
    }
}


NetManager::stream_key::stream_key():
    port(0)
{}


NetManager::stream_key::stream_key(
    string s, string h, u_short p):
    scheme(s), host(h), port(p)
{}


bool
NetManager::stream_key::operator==(
    const stream_key& x) const
{
    return scheme == x.scheme && host == x.host && port == x.port;
}


bool
NetManager::stream_key::operator<(
    const stream_key& x) const
{
    return scheme != x.scheme ? scheme < x.scheme :
        host != x.host ? host < x.host : port < x.port;
}


netstream*
NetManager::get_stream(
    URL url)
{
    stream_key key(url.scheme(), url.host(), url.port());
    alive_streams_type::iterator i = alive_streams_.find(key);
    if (i != alive_streams_.end()) {
        netstream* s = (*i).second;
        alive_streams_.erase(i);
        if (s->is_open()) return s;
    }
    netstream* conn = new netstream(url.host(), url.port());
    if (!conn->is_open()) {
        delete conn;
        conn = NULL;
    }
    return conn;
}


void
NetManager::release_stream(
    netstream* s)
{
#if 0
    if (s->is_open()) {
        alive_streams_type::key_type key(s->host(), s->port());
        alive_streams_[key] = s;
        return;
    }
#endif
    delete s;
}        


void NetManager::add_scheme(string, NetProtocol*) {}
NetProtocol* NetManager::find_scheme(string) {}


string
NetManager::cache_code(
    URL url)
{
    string code = url.scheme() + '/' + url.netloc() + url.path();
    if (code[code.length() - 1] == '/') code += "index.html";
    return code;
}


void
NetManager::cache_dir(
    string dir)
{
    if (dir == cache_dir_) return;
    cache_dir_ = dir;
    if (cache_db_) cache_db_.close();

    string index = cache_dir_ + "/cacheidx.db";
    BTREEINFO info = {
        0,  // flags
        0,  // cachesize
        0,  // maxkeypage
        0,  // minkeypage
        0,  // psize
        cache_compare,
        NULL,
        0,
    };
    cache_db_.open(index.c_str(), O_RDWR | O_CREAT, 0600, DB_BTREE, &info);
}


void
NetManager::register_cache(
    URL url,
    const cache_data& cache)
{
    DBT key;
    key.data = url.abs().c_str();
    key.size = url.abs().length() + 1;

    strstream buf;
    buf << cache << ends;
    DBT data;
    data.data = buf.str();
    data.size = buf.tellp();
    cache_db_.put(&key, &data, 0);
    buf.freeze(false);
}


bool
NetManager::retrieve_cache(
    URL url,
    cache_data& cache)
{
    DBT key;
    key.data = url.abs().c_str();
    key.size = url.abs().length() + 1;

    DBT data;
    if (cache_db_.get(&key, &data, 0) != 0) {
        cache.status = 0;
        cache.code   = string();
        cache.access = cache.expires = 0;
        return false;
    }
    istrstream buf((char*) data.data, data.size);
    buf >> cache;
    return true;
}


int
NetManager::cache_compare(
    const DBT* key1,
    const DBT* key2)
{
    return strcmp((char*) key1->data, (char*) key2->data);
}


istream&
operator>>(
    istream& in,
    NetManager::cache_data& cache)
{
    in >> cache.status >> cache.code >> cache.access >> ws;
    if (in.peek() == '*') cache.expires = (time_t) -1;
    else  in >> cache.expires;
}


ostream&
operator<<(
    ostream& out,
    const NetManager::cache_data& cache)
{
    out << cache.status << ' ' << cache.code << ' ' << cache.access << ' ';
    if (cache.expires == (time_t) -1) out << '*';
    else  out << cache.expires;
    return out;
}
