/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1997-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#ifndef NETMANAGER_H
#define NETMANAGER_H
#ifdef __GNUG__
#pragma interface
#endif

#include "NetMessage.h"
#include "DBinter.h"


class NetRequest
{
 public:
    enum method_t { GET, POST, HEAD };
    enum flags_t  { freeze_bit = 1, nocache_bit = 2, reload_bit = 4 };
    
    explicit NetRequest(URL = URL(), method_t = GET,
                        int flags = 0, NetMessage* = NULL);
    ~NetRequest();

    URL      url()        const { return url_; }
    method_t method()     const { return method_; }
    string   cache_file() const { return cache_file_; }

    NetMessage* message() { return message_; }
    NetHeaders* headers() { return message_->headers(); }
    const NetMessage* message() const { return message_; }
    const NetHeaders* headers() const { return message_->headers(); }

    void url(URL u)             { url_ = u; }
    void method(method_t m)     { method_ = m; }
    void cache_file(string c)   { cache_file_ = c; }
    void message(NetMessage* m) { message_ = m; }
    
    bool is_frozen()    const { return flags_ & freeze_bit; }
    bool is_nocache()   const { return flags_ & nocache_bit; }
    bool is_reloading() const { return flags_ & reload_bit; }
    void set_flags(int f)   { flags_ |= f; }
    void unset_flags(int f) { flags_ &= ~f; }

 private:
    URL         url_;
    method_t    method_;
    NetMessage* message_;
    int         flags_;
    string      cache_file_;

    NetRequest(const NetRequest&);
    NetRequest& operator=(const NetRequest&);
    DBG_NEW
};


class NetResponse
{
 public:
    enum flags_t { freeze_bit = 1, cached_bit = 2, reused_bit = 4 };
    
    NetResponse();
    ~NetResponse();

    URL  url()    const { return url_; }
    int  status() const { return status_; }

    NetMessage* message() { return message_; }
    NetHeaders* headers() { return message_->headers(); }
    const NetMessage* message() const { return message_; }
    const NetHeaders* headers() const { return message_->headers(); }

    void url(URL u)             { url_ = u; }
    void status(int s)          { status_ = s; }
    void message(NetMessage* m) { message_ = m; }

    bool is_frozen() const { return flags_ & freeze_bit; }
    bool is_cached() const { return flags_ & cached_bit; }
    bool is_reused() const { return flags_ & reused_bit; }
    void set_flags(int f)   { flags_ |= f; }
    void unset_flags(int f) { flags_ &= ~f; }
    
 private:
    URL         url_;
    int         status_;
    NetMessage* message_;
    int         flags_;

    NetResponse(const NetResponse&);
    NetResponse& operator=(const NetResponse&);
    DBG_NEW
};


class NetProtocol;
class netstream;

class NetManager
{
 public:
    explicit NetManager(string cache_dir = "");
    virtual ~NetManager();

    void request(NetRequest&, NetResponse&);
    
    netstream* get_stream(URL);
    void       release_stream(netstream*);

    void         add_scheme(string scheme, NetProtocol* proto);
    NetProtocol* find_scheme(string scheme);

    void   cache_dir(string);
    string cache_dir() const { return cache_dir_; }

 protected:
    struct cache_data {
        int    status;
        string code;
        time_t access;
        time_t expires;
    };
    string cache_code(URL);
    void   register_cache(URL, const cache_data&);
    bool   retrieve_cache(URL, cache_data&);
    
 private:
    struct stream_key {
        string  scheme;
        string  host;
        u_short port;
        stream_key();
        stream_key(string, string, u_short);
        bool operator==(const stream_key&) const;
        bool operator< (const stream_key&) const;
    };
    typedef map<stream_key, netstream*, less<stream_key> > alive_streams_type;
    alive_streams_type alive_streams_;

    typedef hmap<string, NetProtocol*> schemes_type;
    schemes_type schemes_;

    string  cache_dir_;
    DBinter cache_db_;
    static int cache_compare(const DBT* key1, const DBT* key2);

    NetProtocol *http_, *file_;
    DBG_NEW
};

istream& operator>>(istream&, NetManager::cache_data&);
ostream& operator<<(ostream&, const NetManager::cache_data&);


class NetProtocol
{
 public:
    NetProtocol(NetManager* manager): manager_(manager) {}
    virtual ~NetProtocol() {}
    
    virtual void get (NetRequest&, NetResponse&) {}
    virtual void post(NetRequest&, NetResponse&) {}
    virtual void head(NetRequest&, NetResponse&) {}

 protected:
    netstream* get_stream(URL url)    { return manager_->get_stream(url); }
    void release_stream(netstream* s) { manager_->release_stream(s); }
    
 private:
    NetManager* manager_;
};


EXTERN NetManager* netmanager;

#endif /* NETNANAGER_H */
