/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#define DBG_MODULE  URL

#ifdef __GNUG__
#pragma implementation
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "URL2.h"

#include <stdlib.h>
#include "util.h"
#include "URL_gen.h"

#if defined __GNUG__ && (defined EGCS || __GNUC__ >= 2 && __GNUC_MINOR__ >= 8)
#define remove erase
#endif


class URL_nil: public URL_rep
{
    friend class URL;

 protected:
    URL_nil();
    virtual ~URL_nil();

    virtual URL_rep* clone() const { return CCAST(URL_nil*, this); }
    
    virtual string scheme()      const { return string(); }
    virtual string parse(int)    const { return string(); }
    virtual string rel(URL, int) const { return string(); }
    virtual bool   is_valid()    const { return false; }

    DBG_NEW
};


URL_nil::URL_nil()
{
    DBG("nil", "creating");
}


URL_nil::~URL_nil()
{
    DBG("nil", "destroying");
}



URL::rep_makers_type rep_makers_;
URL_rep* URL::nil_rep_ = new URL_nil;


URL_rep*
URL::make_rep(
    string url,
    URL    base)
{
    if (url.compare("URL:", 0, 4) == 0) url.remove(0, 4);

    string::size_type i = url.find(':');
    string scheme = i != string::npos ? url.substr(0, i): string();
    if (scheme != "http" && scheme != "ftp" &&
        scheme != "file" && scheme != "mailto")
        scheme = base.scheme();

    URL_rep* rep = NULL;
    if      (scheme == "http")   rep = URL_http::make_rep(url, base);
    else if (scheme == "ftp")    rep = URL_ftp ::make_rep(url, base);
    else if (scheme == "file")   rep = URL_file::make_rep(url, base);
    // else if (scheme == "mailto") rep = URL_mailto::make_rep(url, base);
    if (!rep) rep = nil_rep_;

    return rep;
}


void
URL::crack_url(
    string  url,
    string& scheme,
    string& netloc,
    string& path,
    string& query,
    string& params,
    string& frag)
{
    scheme = netloc = path = query = params = frag = string();
    if (url.compare("URL:", 0, 4) == 0) url.remove(0, 4);

    string::size_type sharp  = url.rfind('#');
    if (sharp != string::npos) {
        frag = url.substr(sharp + 1);
        url.remove(sharp);
    }

    string::size_type qmark  = url.rfind('?');
    if (qmark != string::npos) {
        query = url.substr(qmark + 1);
        url.remove(qmark);
    }

    string::size_type scolon = url.rfind(';');
    if (scolon != string::npos) {
        params = url.substr(scolon + 1);
        url.remove(scolon);
    }

    string::size_type slasla = url.find("//");
    if (slasla != string::npos) {
        netloc = url.substr(slasla + 2);
        string::size_type slash = netloc.find('/');
        if (slash != string::npos) netloc.remove(slash);
        url.remove(slasla, netloc.length() + 2);
    }

    string::size_type colon = url.find(':');
    if (colon != string::npos) {
        scheme = url.substr(0, colon);
        url.remove(0, colon + 1);
    }

    path = url;
}


void
URL::crack_netloc(
    string   netloc,
    string&  username,
    string&  password,
    string&  hostname,
    u_short& port)
{
    username = password = hostname = string();
    port = 0;

    string::size_type atmark = netloc.find('@');
    if (atmark != string::npos) {
        string::size_type colon1 = netloc.rfind(':', atmark);
        if (colon1 != string::npos) {
            username = netloc.substr(0, colon1);
            password = netloc.substr(colon1+1, atmark - colon1-1);
        }
        else {
            username = netloc.substr(0, atmark);
        }
        netloc.remove(0, atmark + 1);
    }

    string::size_type colon2 = netloc.find(':');
    if (colon2 != string::npos) {
        port = strtoul(netloc.substr(colon2 + 1).c_str(), NULL, 10);
        netloc.remove(colon2);
    }

    hostname = netloc;
}

    
string
URL::escape(
    string  str,
    field_t field)
{
    int mode =
        field == path_bit  ? 2:
        field == query_bit ? 4:
        field == frag_bit  ? 4: 1;
    
    // 1: safe always
    // 2: safe in path + '~' (safe + '/' + '~')
    // 4: safe in query, params and frag (safe + '&' + '=' + ':' + '/' + '~')
    static char is_safe[] =
    { //0 1 2 3 4 5 6 7 8 9 A B C D E F
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // 0
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // 1
        0,0,0,0,0,0,4,0,0,0,7,7,0,7,7,6,  // 2   !"#$%&'()*+,-./     // "
        7,7,7,7,7,7,7,7,7,7,4,0,0,4,0,0,  // 3  0123456789:;<=>?
        7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,  // 3  @ABCDEFGHIJKLMNO
        7,7,7,7,7,7,7,7,7,7,7,0,0,0,0,7,  // 4  PQRSTUVWXYZ[\]^_
        0,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,  // 6  `abcdefghijklmno
        7,7,7,7,7,7,7,7,7,7,7,0,0,0,6,0,  // 7  pqrstuvwxyz{\}~
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // 8
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // 9
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // A
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // B
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // C
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // D
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // E
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  // F
    };

    static char *hex = "0123456789ABCDEF";

    string result;
    for (string::size_type i = 0, len = str.length(); i < len; i++) {
        unsigned char c = str[i];
        if (is_safe[c] & mode) {
            result += c;
        }
        else {
            result += '%';
            result += hex[c >> 4];
            result += hex[c & 0x0f];
        }
    }
    return result;
}


string
URL::unescape(
    string str)
{
    string::size_type i = str.find('%');
    if (i == string::npos) return str;
    
    string::size_type len = str.length();
    string result(str, 0, i);
    result.resize(len);

    while (i < len) {
        if (str[i] != '%') {
            result += str[i++];
        }
        else {
            i++;
            if (i >= len) break;
            unsigned char hi = str[i++];
            if      ('0' <= hi && hi <= '9') hi -= '0';
            else if ('A' <= hi && hi <= 'F') hi -= 'A';
            else if ('a' <= hi && hi <= 'f') hi -= 'a';
            else { result += '%';  i--;  continue; }
            hi <<= 4;

            if (i >= len) break;
            unsigned char lo = str[i++];
            if      ('0' <= lo && lo <= '9') lo -= '0';
            else if ('A' <= lo && lo <= 'F') lo -= 'A';
            else if ('a' <= lo && lo <= 'f') lo -= 'a';
            else { result += '%';  i -= 2;  continue; }

            result += hi | lo;
        }
    }
    return result;

#if 0
    for (string::size_type i = 0;
         (i = str.find('%', i)) != string::npos && i + 2 < str.length(); i++) {

        unsigned char hi = str[i + 1];
        if      ('0' <= hi && hi <= '9') hi -= '0';
        else if ('A' <= hi && hi <= 'F') hi -= 'A';
        else if ('a' <= hi && hi <= 'f') hi -= 'a';
        else  continue;
        hi <<= 4;

        if (i >= len) break;
        unsigned char lo = str[i + 2];
        if      ('0' <= lo && lo <= '9') lo -= '0';
        else if ('A' <= lo && lo <= 'F') lo -= 'A';
        else if ('a' <= lo && lo <= 'f') lo -= 'a';
        else  continue;

        str.replace(i, 3, hi | lo);
    }
    retrun str;
#endif
}


URL::URL():
    refcnt(nil_rep_)
{}


URL::URL(
    const URL& url):
    refcnt(url.rep())
{}


URL::URL(
    string url):
    refcnt(URL::make_rep(url, URL()))
{}


URL::URL(
    string url,
    URL    base):
    refcnt(URL::make_rep(url, base))
{}


URL::~URL()
{}    


URL&
URL::operator=(
    const URL& url)
{
    this->set_rep(url.rep());
    return *this;
}


URL&
URL::operator=(
    string url)
{
    this->set_rep(URL::make_rep(url, *this));
    return *this;
}


bool
URL::scheme(
    string s)
{
    if (s == this->scheme()) return true;
    this->set_rep(URL::make_rep(s + ":", *this));
    return true;
}


#define DEFINE_SET_ATTR(name, type)          \
bool                                         \
URL::name(                                   \
    type value)                              \
{                                            \
    if (value == this->name()) return true;  \
    bool f = false;                          \
                                             \
    if (rep()->ref() > 1) {                  \
        URL_rep* r = rep()->clone();         \
        if (!r) return false;                \
	f = r->name(value);                  \
        this->set_rep(r);                    \
    }                                        \
    else {                                   \
	f = rep()->name(value);              \
	rep()->regrab();                     \
    }                                        \
    return f;                                \
}

DEFINE_SET_ATTR(host,     string)
DEFINE_SET_ATTR(user,     string)
DEFINE_SET_ATTR(password, string)
DEFINE_SET_ATTR(port,     u_short)
DEFINE_SET_ATTR(path,     string)
DEFINE_SET_ATTR(query,    string)
DEFINE_SET_ATTR(params,   string)
DEFINE_SET_ATTR(frag,     string)
DEFINE_SET_ATTR(equery,   string)
DEFINE_SET_ATTR(eparams,  string)
DEFINE_SET_ATTR(efrag,    string)


ostream&
operator<<(
    ostream& out,
    URL      url)
{
    return out << url.abs();
}


istream&
operator>>(
    istream& in,
    URL&     url)
{
    string s;
    in >> s;
    url = s;
    return in;
}
