/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1996-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#ifndef URL_H
#define URL_H
#ifdef __GNUG__
#pragma interface
#endif

#include <sys/types.h>
#include <istream.h>
#include <ostream.h>
#include "utilstl.h"
#ifdef __GNUG__
#define FUNCTION_H
#endif
#include <string>
#include "refcnt.h"

class URL;


class URL_rep: public refcnt_rep
{
    friend class URL;

 protected:
    URL_rep() {}
    virtual ~URL_rep() {}

    virtual URL_rep* clone()    const = 0;
    virtual bool     is_valid() const = 0;

    // generic URL
    virtual string  scheme()   const = 0;
    virtual string  host()     const { return string(); }
    virtual string  user()     const { return string(); }
    virtual string  password() const { return string(); }
    virtual u_short port()     const { return 0; }
    virtual string  path()     const { return string(); }
    virtual string  query()    const { return string(); }
    virtual string  params()   const { return string(); }
    virtual string  frag()     const { return string(); }
    virtual string  netloc()   const { return string(); }
    virtual string  equery()   const { return string(); }
    virtual string  eparams()  const { return string(); }
    virtual string  efrag()    const { return string(); }

    virtual bool host(string)     { return false; }
    virtual bool user(string)     { return false; }
    virtual bool password(string) { return false; }
    virtual bool port(u_short)    { return false; }
    virtual bool path(string)     { return false; }
    virtual bool query(string)    { return false; }
    virtual bool params(string)   { return false; }
    virtual bool frag(string)     { return false; }
    virtual bool equery(string)   { return false; }
    virtual bool eparams(string)  { return false; }
    virtual bool efrag(string)    { return false; }

    // news:
    virtual string messageid() const { return string(); }
    virtual string group()     const { return string(); }
    
    enum field_t {
	scheme_bit =  1,  // scheme:
	netloc_bit =  2,  // //user:password@host:port
	path_bit   =  4,  // /path...
        query_bit  =  8,  // ?query;params
	frag_bit   = 16,  // #frag
        all        = 31,
    };

    string         abs()         const { return abs_; }
    virtual string parse(int)    const = 0;
    virtual string rel(URL, int) const = 0;
    virtual string reqpath()     const { return parse(path_bit | query_bit); }
    virtual string basename()    const { return string(); }

    virtual void grab() { if (!ref()) abs_ = parse(31); refcnt_rep::grab(); }
    virtual void regrab() { abs_ = parse(31); }
    
 private:
    string abs_;
    DBG_NEW
};


//////////////////////////////////////////////////////////////////////////////
//  URL -- canonicalization and string conversion of URL(URI)
//
class URL: public refcnt
{
 public:
    // primitive functions
    URL();
    URL(const URL&);
    URL(string url);
    URL(string url, URL base);
    ~URL();

    URL& operator=(const URL&);
    URL& operator=(string);
    
    // generic URL
    string  scheme()   const { return rep()->scheme(); }
    string  host()     const { return rep()->host(); }
    string  user()     const { return rep()->user(); }
    string  password() const { return rep()->password(); }
    u_short port()     const { return rep()->port(); }
    string  path()     const { return rep()->path(); }
    string  query()    const { return rep()->query(); }
    string  params()   const { return rep()->params(); }
    string  frag()     const { return rep()->frag(); }
    string  netloc()   const { return rep()->netloc(); }   // escaped netloc
    string  equery()   const { return rep()->equery(); }   // escaped query
    string  eparams()  const { return rep()->eparams(); }  // escaped params
    string  efrag()    const { return rep()->efrag(); }    // escaped frag

    bool scheme(string);
    bool host(string);
    bool user(string);
    bool password(string);
    bool port(u_short);
    bool path(string);
    bool query(string);
    bool params(string);
    bool frag(string);
    bool equery(string);
    bool eparams(string);
    bool efrag(string);

    typedef URL_rep::field_t field_t;
    enum {
	scheme_bit = URL_rep::scheme_bit,  // scheme:
	netloc_bit = URL_rep::netloc_bit,  // //user:password@host:port
	path_bit   = URL_rep::path_bit,    // /path...
        query_bit  = URL_rep::query_bit,   // ?query;params
	frag_bit   = URL_rep::frag_bit,    // #frag
        all        = URL_rep::all,
    };
    string abs()                       const { return rep()->abs(); }
    string parse(int bits = all)       const { return rep()->parse(bits); }
    string rel(URL base, int bits=all) const { return rep()->rel(base, bits); }
    string reqpath()                   const { return rep()->reqpath(); }
    string basename()                  const { return rep()->basename(); }

    bool is_valid() const { return rep()->is_valid(); }
    operator bool() const { return is_valid(); }

    bool operator==(URL x) const { return rep()==x.rep() || abs()==x.abs(); }
    bool operator< (URL x) const { return rep()< x.rep() || abs()< x.abs(); }
    
    // scheme://netloc/path?query;params#frag
    static void crack_url(string url, string& s, string& n, string& p,
                          string& q, string& p, string& f);
    // username:password@hostname:port
    static void crack_netloc(string netloc,
                             string& u, string& p, string& h, u_short& p);

    static string escape(string, field_t);
    static string unescape(string);

 private:
    virtual       URL_rep* rep() {  return DCAST(URL_rep*, refcnt::rep()); }
    virtual const URL_rep* rep() const {
        return DCAST(const URL_rep*, refcnt::rep()); }
    
    typedef URL_rep* (*rep_maker)(string, URL);
    typedef hmap<string, rep_maker> rep_makers_type;
    static rep_makers_type rep_makers_;
    static URL_rep* make_rep(string, URL);
    static URL_rep* nil_rep_;
    DBG_NEW
};

ostream& operator<<(ostream&, URL);
istream& operator>>(istream&, URL&);

#endif  /* URL_H */
