/*                                               -*- Mode:C++; tab-width: 8 -*-
//  The Natto View
//
//  Copyright (c) 1995-1998  Hidekazu Shiozawa <shiozawa@myo.inst.keio.ac.jp>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU *Library* General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
*/

#ifdef __GNUG__
#pragma implementation
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#define NO_MSGOUT
#include "util.h"

#include <ctype.h>
#include <istream.h>
#include <ostream.h>
#include <vector>

#if defined __GNUG__ && (defined EGCS || __GNUC__ >= 2 && __GNUC_MINOR__ >= 8)
#define remove erase
#endif


static coutbuf coutbuf_instance;
ostream msgout(&coutbuf_instance);


istream&
copy_line(
    istream& in,
    ostream& out)
{
    const unsigned char CR = '\015';
    const unsigned char LF = '\012';
    
    unsigned char ch = '\0';
    while (in.get(ch) && (ch == ' ' || ch == '\t'))
	;
    switch (ch) {
     case CR:
	if (in.peek() == LF) (void) in.get(ch);
	return in;
     case LF:
     case '\0':
	return in;
    }

    streamoff n = 0;
    do { 
	if (isspace(ch)) n++;
	else n = 0;
	out.put(ch);
	ch = '\0';
    } while (in.get(ch) && ch != CR && ch != LF && ch != '\0');
    out.seekp(-n, ios::cur);

    if (ch == CR && in.peek() == LF) (void) in.get(ch);
    return in;
}


string
trim_space(
    string str)
{
    string::size_type len = str.length();
    string::size_type i = 0;
    for (i = 0; i < len && isspace(str[len-1-i]); i++)
        ;
    str.remove(len - i, i);

    len = str.length();
    for (i = 0; i < len && isspace(str[i]); i++)
        ;
    str.remove(0, i);

    return str;
}


string
squeeze_space(
    string str)
{
    str = trim_space(str);

    string::size_type len = str.length();
    string::size_type i = 0;
    while (i < len) {
        while (!isspace(str[i]))
            if (++i >= len) return str;
        string::size_type j = i++;
        while (isspace(str[i]))
            i++;
        str.replace(j, i - j, ' ');
    }
    return str;
}


string
lc(
    string str)
{
    for (string::size_type i = str.length() - 1;
         i != (string::size_type) -1; i--)
        str[i] = tolower(str[i]);
    return str;
}


string
uc(
    string str)
{
    for (string::size_type i = str.length() - 1;
         i != (string::size_type) -1; i--)
        str[i] = toupper(str[i]);
    return str;
}


string
capitalize2(
    string str)
{
    string::size_type len = str.length();
    bool prev_alnum = false;
    for (string::size_type i = 0; i < len; i++) {
        if (isalnum(str[i])) {
            str[i] = prev_alnum ? tolower(str[i]): toupper(str[i]);
            prev_alnum = true;
        }
        else {
            prev_alnum = false;
        }
    }
    return str;
}


string
simplify_path(
    string path)
{
    bool is_full = path[0] == '/';
    bool is_dir  = path.length() && path[path.length() - 1] == '/';
    
    typedef vector<string> strvector;
    strvector vec1;

    back_insert_iterator<strvector> ins(vec1);
    split(path, ins, string("/"));

    strvector vec2;
    for (strvector::iterator i = vec1.begin(); i != vec1.end(); i++) {
        if (*i == "." || *i == "") continue;
        if (*i == "..") {
            if (vec2.size()) { vec2.pop_back(); continue; }
            else if (is_full) continue;
        }
        vec2.push_back(*i);
    }
    path = join(vec2.begin(), vec2.end(), string("/"));

    if (is_full) path = '/' + path;
    if (is_dir && path.length() && path[path.length() - 1] != '/')
        path += '/';
    return path;
}


string
basename(
    string path,
    string ext = "")
{
    string::size_type i, len = path.length();
    if (!len) return path;

    for (i = 0; i < len && path[len - 1 - i] == '/'; i++)
        ;
    path.remove(len - i, i);

    string::size_type pos = path.rfind('/');
    if (pos != string::npos) path.remove(0, pos + 1);
    pos = path.rfind(ext);
    if (pos == path.length() - ext.length() + 1)
        path.remove(pos);
    return path;
}


string
dirname(
    string path)
{
    if (path == "/") return path;
    string::size_type i, len = path.length();
    if (!len) return path;
    
    for (i = 0; i < len && path[len - 1 - i] != '/'; i++)
        ;
    path.remove(len - i, i);

    string::size_type end = path.rfind('/');
    switch (end) {
     case 0:            return "/";
     case string::npos: return ".";
     default:           path.remove(end);  return path;
    }
}


string
absolute_path(
    string relpath,
    string curpath)
{
    relpath = simplify_path(relpath);
    curpath = simplify_path(curpath);
    if (relpath[0] == '/') return relpath;
    if (curpath.length() && curpath[curpath.length() - 1] != '/')
        curpath = dirname(curpath) + '/';
    return simplify_path(curpath + relpath);
}


string
relative_path(
    string abspath,
    string curpath)
{
    abspath = simplify_path(abspath);
    curpath = simplify_path(curpath);
    if (!abspath.length() || abspath[0] != '/' ||
        !curpath.length() || curpath[0] != '/')
        return simplify_path(abspath);
    if (curpath.length() && curpath[curpath.length() - 1] != '/')
        curpath = dirname(curpath) + '/';

    typedef vector<string> strvector;
    strvector::size_type i = 0, slash = 0;
    strvector::size_type size = min(abspath.length(), curpath.length());
    for (i = 0; i < size && abspath[i] == curpath[i]; i++)
        if (abspath[i] == '/') slash = i;
    if (!slash) return simplify_path(abspath);

    string result;
    for (i = slash + 1;;) {
        i = curpath.find('/', i);
        if (i == string::npos) break;
        result += "../";
    }
    result += abspath.substr(i + 1);
    return result;
}


int
coutbuf::sync()
{
    streamsize n = this->pptr() - this->pbase();
    cout.write(this->pbase(), n);
    cout.flush();
    return 0;
}


int
coutbuf::overflow(
    int ch)
{
    streamsize n = this->pptr() - this->pbase();
    if (n && this->sync() != 0) return EOF;
    if (ch != EOF) cout.put((char) ch);
    this->pbump(-n);
    return 0;
}


streamsize
coutbuf::xsputn(
    char*      text,
    streamsize n)
{
    if (this->sync() == EOF) return 0;
    cout.write(text, n);
    return n;
}
